import datetime
import os
import pytest
import shlex

from ligo.gracedb.cli.client import CommandLineInterface
from ligo.gracedb.exceptions import HTTPError

# Apply module-level marks
pytestmark = [pytest.mark.cli, pytest.mark.integration]


# Instantiated client for use with tests
CLI = CommandLineInterface()

# Get test service URL and data directory
TEST_SERVICE = os.environ.get(
    'TEST_SERVICE',
    'https://gracedb-test.ligo.org/api/'
)
TEST_DATA_DIR = os.environ.get(
    'TEST_DATA_DIR',
    os.path.join(os.path.dirname(__file__), 'data')
)


# Utility function for running command with CLI and specifying the test service
def run_CLI_test(cmd):
    cmd += ' --service-url={url}'.format(url=TEST_SERVICE)
    return CLI(shlex.split(cmd))


###############################################################################
# Tests #######################################################################
###############################################################################
# Order of tests matters to some extent - we test operations first which are
# used to set up later operations. (Ex: we test creating a event before we
# use event creation in the test for getting an event)
def test_create_event():
    """Test creating an event with the command-line interface"""
    # Set up command
    cmd = 'create event Test gstlal {event_file} --labels=INJ,DQV --offline' \
        .format(event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))

    # Make request and get response
    response = run_CLI_test(cmd)
    out = response.json()

    # Check output
    assert response.status_code == 201
    assert out['group'] == 'Test'
    assert out['pipeline'] == 'gstlal'
    assert out['search'] is None
    assert sorted(out['labels']) == ['DQV', 'INJ']
    assert out['offline'] is True


def test_get_event():
    """Test getting an event with the command-line interface"""
    # Setup: create event
    cmd = 'create event Test gstlal {event_file} LowMass '.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response = run_CLI_test(cmd)
    gid = response.json()['graceid']

    # Get event
    cmd = 'get event {gid}'.format(gid=gid)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check data
    assert response.status_code == 200
    assert out['group'] == 'Test'
    assert out['pipeline'] == 'gstlal'
    assert out['search'] == 'LowMass'
    assert sorted(out['labels']) == []
    assert out['offline'] is False


def test_update_event():
    """Test updating an event with the CLI"""
    # Setup: create event
    cmd = 'create event Test gstlal {event_file} LowMass '.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response = run_CLI_test(cmd)
    assert response.status_code == 201
    out = response.json()
    gid = out['graceid']

    # Update event
    filename = os.path.join(TEST_DATA_DIR, 'cbc-lm2.xml')
    cmd = 'update event {gid} {filename}'.format(gid=gid, filename=filename)
    response = run_CLI_test(cmd)

    # Check response
    assert response.status_code == 202

    # Replace event response is useless so we get the event and check it
    cmd = 'get event {gid}'.format(gid=gid)
    response = run_CLI_test(cmd)
    out3 = response.json()

    # Check data
    assert response.status_code == 200
    assert out3['graceid'] == out['graceid']
    assert out3['gpstime'] != out['gpstime']  # gpstime should have changed


def test_search_events():
    """Test searching for events with the CLI"""
    # Setup: create an event
    cmd = 'create event Test gstlal {event_file}'.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response = run_CLI_test(cmd)
    assert response.status_code == 201
    out = response.json()
    gid = out['graceid']

    # Search events -----------------------------------------------------------
    cmd = 'search events {gid}'.format(gid=gid)
    response = run_CLI_test(cmd)

    # Check response (tab delimited)
    lines = response.split('\n')
    content = lines[1].split('\t')
    assert len(lines) == 2  # 1 header, 1 actual content
    assert content[0] == gid
    assert content[1] == ''  # labels
    assert content[2] == out['group']
    assert content[3] == out['pipeline']
    assert content[4] == str(out['far'])
    assert content[5] == str(out['gpstime'])
    assert content[6] == out['created']


def test_labels():
    """Test all things label related with the CLI"""
    # Setup: create event
    cmd = 'create event Test gstlal {event_file} LowMass '.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response = run_CLI_test(cmd)
    gid = response.json()['graceid']

    # Add a label -------------------------------------------------------------
    label = 'PE_READY'
    cmd = 'add label {gid} {label}'.format(gid=gid, label=label)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check response
    assert response.status_code == 201

    # Get event and check
    cmd = 'get event {gid}'.format(gid=gid)
    response = run_CLI_test(cmd)
    out = response.json()
    assert response.status_code == 200
    assert out['labels'] == [label]

    # Get label ---------------------------------------------------------------
    cmd = 'get label {gid} {label}'.format(gid=gid, label=label)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check response
    assert response.status_code == 200
    assert out['name'] == label

    # List labels -------------------------------------------------------------
    cmd = 'list labels {gid}'.format(gid=gid)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check response
    assert response.status_code == 200
    assert len(out['labels']) == 1
    assert out['labels'][0]['name'] == label

    # Remove label ------------------------------------------------------------
    cmd = 'remove label {gid} {label}'.format(gid=gid, label=label)
    response = run_CLI_test(cmd)

    # Check response
    assert response.status_code == 204

    # Get event and check
    cmd = 'get event {gid}'.format(gid=gid)
    response = run_CLI_test(cmd)
    out = response.json()
    assert response.status_code == 200
    assert out['labels'] == []


def test_logs_files():
    """Test all things log and file related with CLI"""
    # Setup: create event
    cmd = 'create event Test gstlal {event_file} LowMass '.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response = run_CLI_test(cmd)
    gid = response.json()['graceid']

    # Create a log message with a file ----------------------------------------
    comment = "test comment"
    fname = 'test_file.txt'
    filename = os.path.join(TEST_DATA_DIR, fname)
    tags = ['tag1', 'tag2']
    cmd = 'create log {gid} "{comment}" {filename} --tag-name={tags}'.format(
        gid=gid, comment=comment, filename=filename, tags=",".join(tags))
    response = run_CLI_test(cmd)
    out = response.json()

    # Test output
    assert response.status_code == 201
    assert out['comment'] == out['comment']
    assert out['filename'] == os.path.basename(filename)
    assert sorted(out['tag_names']) == sorted(tags)

    # Get log -----------------------------------------------------------------
    cmd = 'get log {gid} {N}'.format(gid=gid, N=out['N'])
    response = run_CLI_test(cmd)
    out = response.json()

    # Test output
    assert response.status_code == 200
    assert out['comment'] == out['comment']
    assert out['filename'] == fname
    assert sorted(out['tag_names']) == sorted(tags)

    # List logs ---------------------------------------------------------------
    cmd = 'list logs {gid}'.format(gid=gid)
    response = run_CLI_test(cmd)
    out = response.json()

    # Test output
    assert response.status_code == 200
    comments = [l['comment'] for l in out['log']]
    assert comment in comments

    # Get file ----------------------------------------------------------------
    cmd = 'get file {gid} {filename}'.format(gid=gid, filename=fname)
    response = run_CLI_test(cmd)
    with open(filename, 'rb') as fh:
        assert response == fh.read()

    # List files -------------------------------------------------------------
    cmd = 'list files {gid}'.format(gid=gid, filename=fname)
    response = run_CLI_test(cmd)
    out = response.json()
    assert response.status_code == 200
    assert len(out) >= 2
    assert fname in out
    assert fname + ',0' in out


def test_tags():
    """Test all things tag-related with the CLI"""
    # Setup: create event
    cmd = 'create event Test gstlal {event_file} LowMass '.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response = run_CLI_test(cmd)
    gid = response.json()['graceid']

    # Setup: create a log message
    comment = "test comment"
    filename = os.path.join(TEST_DATA_DIR, 'test_file.txt')
    tags = ['tag1', 'tag2']
    cmd = 'create log {gid} "{comment}" {filename} --tag-name={tags}'.format(
        gid=gid, comment=comment, filename=filename, tags=",".join(tags))
    out = run_CLI_test(cmd).json()
    log_N = out['N']
    assert response.status_code == 201
    assert out['comment'] == out['comment']
    assert out['filename'] == os.path.basename(filename)
    assert sorted(out['tag_names']) == sorted(tags)

    # Add tag to log ----------------------------------------------------------
    tag = 'tag3'
    cmd = 'add tag {gid} {N} {tag}'.format(gid=gid, N=log_N, tag=tag)
    response = run_CLI_test(cmd)
    assert response.status_code == 201

    # Tag response for events is not helpful. So we have to get the full log
    cmd = 'get log {gid} {N}'.format(gid=gid, N=log_N)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check output
    assert response.status_code == 200
    assert sorted(out['tag_names']) == sorted(tags + [tag])

    # List tags ---------------------------------------------------------------
    cmd = 'list tags {gid} {N}'.format(gid=gid, N=log_N)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check response
    assert response.status_code == 200
    assert len(out['tags']) == 3
    tags_list = [t['name'] for t in out['tags']]
    assert sorted(tags_list) == sorted(tags + [tag])

    # Remove tag from log -----------------------------------------------------
    cmd = 'remove tag {gid} {N} {tag}'.format(gid=gid, N=log_N, tag=tag)
    response = run_CLI_test(cmd)
    assert response.status_code == 204

    # Tag response for events is not helpful. So we have to get the full log
    cmd = 'get log {gid} {N}'.format(gid=gid, N=log_N)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check output
    assert response.status_code == 200
    assert sorted(out['tag_names']) == sorted(tags)
    assert tag not in out['tag_names']


def test_create_superevent():
    """Test creating a superevent with the command-line interface"""
    # Setup: create two events
    cmd = 'create event Test gstlal {event_file}'.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response1 = run_CLI_test(cmd)
    gid1 = response1.json()['graceid']
    response2 = run_CLI_test(cmd)
    gid2 = response2.json()['graceid']

    # Create superevent
    cmd = ('create superevent 1 2 3 {p_ev} --labels=INJ,EM_READY '
           '--category=Test --events={ev}').format(p_ev=gid1, ev=gid2)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check response content
    assert response.status_code == 201
    assert out['category'] == 'Test'
    assert sorted(out['gw_events']) == sorted([gid1, gid2])
    assert sorted(out['labels']) == ['EM_READY', 'INJ']
    assert out['preferred_event'] == gid1
    assert out['t_start'] == 1
    assert out['t_0'] == 2
    assert out['t_end'] == 3
    assert out['em_events'] == []


def test_get_superevent():
    """Test getting a superevent with the command-line interface"""
    # Setup: create event
    cmd = 'create event Test gstlal {event_file}'.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response = run_CLI_test(cmd)
    gid = response.json()['graceid']

    # Setup: create superevent
    cmd = 'create superevent 1 2 3 {p_ev} --category=Test '.format(p_ev=gid)
    response = run_CLI_test(cmd)
    out = response.json()
    sid = out['superevent_id']

    # Get superevent
    cmd = 'get superevent {sid}'.format(sid=sid)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check response
    assert out['preferred_event'] == gid
    assert out['gw_events'] == [gid]
    assert out['em_events'] == []
    assert out['category'] == 'Test'
    assert out['t_start'] == 1
    assert out['t_0'] == 2
    assert out['t_end'] == 3


def test_add_event_to_superevent():
    """Test adding an event to a superevent with the command-line interface"""
    # Setup: create two events
    cmd = 'create event Test gstlal {event_file}'.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response1 = run_CLI_test(cmd)
    gid1 = response1.json()['graceid']
    response2 = run_CLI_test(cmd)
    gid2 = response2.json()['graceid']

    # Setup: create superevent with event 1
    cmd = 'create superevent 1 2 3 {p_ev} --category=Test '.format(p_ev=gid1)
    response = run_CLI_test(cmd)
    out = response.json()
    sid = out['superevent_id']

    # Check contents
    assert response.status_code == 201
    assert out['category'] == 'Test'
    assert out['preferred_event'] == gid1
    assert out['gw_events'] == [gid1]

    # Add event 2 to superevent
    cmd = 'add event {sid} {gid}'.format(sid=sid, gid=gid2)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check response
    assert out['graceid'] == gid2

    # Get superevent and check more details
    cmd = 'get superevent {sid}'.format(sid=sid)
    response = run_CLI_test(cmd)
    out = response.json()
    assert sorted(out['gw_events']) == sorted([gid1, gid2])
    assert out['preferred_event'] == gid1


def test_remove_event_from_superevent():
    """Test adding an event to a superevent with the command-line interface"""
    # Setup: create two events
    cmd = 'create event Test gstlal {event_file}'.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response1 = run_CLI_test(cmd)
    gid1 = response1.json()['graceid']
    response2 = run_CLI_test(cmd)
    gid2 = response2.json()['graceid']

    # Setup: create superevent with both events
    cmd = ('create superevent 1 2 3 {p_ev} --category=Test --events={ev}') \
        .format(p_ev=gid1, ev=gid2)
    response = run_CLI_test(cmd)
    out = response.json()
    sid = out['superevent_id']

    # Check contents
    assert response.status_code == 201
    assert out['category'] == 'Test'
    assert out['preferred_event'] == gid1
    assert sorted(out['gw_events']) == sorted([gid1, gid2])

    # Remove event 2 from superevent
    cmd = 'remove event {sid} {gid}'.format(sid=sid, gid=gid2)
    response = run_CLI_test(cmd)

    # Check response
    assert response.status_code == 204

    # Get superevent and make sure event 2 is not in it
    cmd = 'get superevent {sid}'.format(sid=sid)
    response = run_CLI_test(cmd)
    out = response.json()
    assert out['gw_events'] == [gid1]
    assert out['preferred_event'] == gid1


def test_update_superevent():
    """Test update superevent with CLI"""
    # Setup: create two events
    cmd = 'create event Test gstlal {event_file}'.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response1 = run_CLI_test(cmd)
    assert response1.status_code == 201
    gid1 = response1.json()['graceid']
    response2 = run_CLI_test(cmd)
    assert response2.status_code == 201
    gid2 = response2.json()['graceid']

    # Setup: create superevent
    cmd = ('create superevent 1 2 3 {p_ev} --labels=INJ,EM_READY '
           '--category=Test').format(p_ev=gid1)
    response = run_CLI_test(cmd)
    out = response.json()
    assert response.status_code == 201
    sid = out['superevent_id']

    # Update superevent -------------------------------------------------------
    cmd = ('update superevent {sid} --t-start=4 --t-0=5 --t-end=6 '
           '--preferred-event={gid}').format(sid=sid, gid=gid2)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check response
    assert response.status_code == 200
    assert out['t_start'] == 4
    assert out['t_0'] == 5
    assert out['t_end'] == 6
    assert out['preferred_event'] == gid2
    assert sorted(out['gw_events']) == sorted([gid1, gid2])


def test_update_grbevent():
    """Test update GRB event with CLI"""
    # Setup: create a GRB event
    cmd = 'create event Test Fermi {event_file} GRB'.format(
        event_file=os.path.join(TEST_DATA_DIR, 'fermi-test.xml'))
    response = run_CLI_test(cmd)
    assert response.status_code == 201
    initial_data = response.json()
    gid = initial_data['graceid']

    # Update the grbevent's parameters
    redshift = 3.4
    designation = 'very good'
    ra = 12.34
    cmd = ('update grbevent {gid} --redshift={rs} --designation="{des}" '
           '--ra={ra}').format(gid=gid, rs=redshift, des=designation,
                               ra=ra)
    try:
        response = run_CLI_test(cmd)
        new_data = response.json()

        # Even though they're test GRB events, unprivileged users
        # can't update GRB events. So try to catch a 200 status
        # in the case of an admin or grb user, and 403 otherwise.
        # This came up with setting up the gitlab integration instance.

        assert response.status_code == 200
        # Compare results
        initial_grb_params = initial_data['extra_attributes']['GRB']
        new_grb_params = new_data['extra_attributes']['GRB']
        assert new_grb_params['ra'] == ra
        assert new_grb_params['redshift'] == redshift
        assert new_grb_params['designation'] == designation
        assert new_grb_params['ra'] != initial_grb_params['ra']
        assert new_grb_params['redshift'] != initial_grb_params['redshift']

    except HTTPError as e:
        assert e.status_code == 403


def test_confirm_as_gw():
    """Test confirming a superevent as a GW with the command-line interface"""
    # Setup: create an event
    cmd = 'create event Test gstlal {event_file}'.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response = run_CLI_test(cmd)
    gid = response.json()['graceid']

    # Setup: create superevent
    cmd = 'create superevent 1 2 3 {p_ev} --category=Test'.format(p_ev=gid)
    response = run_CLI_test(cmd)
    out = response.json()
    sid = out['superevent_id']
    assert response.status_code == 201
    assert out['gw_id'] is None

    # Confirm as GW
    cmd = 'confirm_as_gw {sid}'.format(sid=sid)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check output
    assert response.status_code == 200
    assert out['superevent_id'] == sid
    assert out['gw_id'] is not None
    assert out['gw_id'].startswith('TGW')


def test_search_superevents():
    """Test searching for superevents with the CLI"""
    # Setup: create an event
    cmd = 'create event Test gstlal {event_file}'.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response = run_CLI_test(cmd)
    assert response.status_code == 201
    gid = response.json()['graceid']

    # Setup: create superevent
    cmd = ('create superevent 1 2 3 {p_ev} --labels=INJ,EM_READY '
           '--category=Test').format(p_ev=gid)
    response = run_CLI_test(cmd)
    out = response.json()
    assert response.status_code == 201
    sid = out['superevent_id']

    # Search superevents ------------------------------------------------------
    cmd = 'search superevents {sid}'.format(sid=sid)
    response = run_CLI_test(cmd)

    # Check response (tab delimited)
    lines = response.split('\n')
    content = lines[1].split('\t')
    assert len(lines) == 2  # 1 header, 1 actual content
    assert content[0] == sid
    assert content[1] == gid
    assert content[2] == gid
    assert sorted(content[3].split(',')) == sorted(out['labels'])
    assert content[4] == str(out['far'])
    assert content[5] == out['links']['files']


def test_permissions():
    """Test exposing and hiding a superevent with the CLI"""
    # Setup: create an event
    cmd = 'create event Test gstlal {event_file}'.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response = run_CLI_test(cmd)
    assert response.status_code == 201
    gid = response.json()['graceid']

    # Setup: create superevent
    cmd = 'create superevent 1 2 3 {p_ev} --category=Test'.format(p_ev=gid)
    response = run_CLI_test(cmd)
    out = response.json()
    assert response.status_code == 201
    sid = out['superevent_id']

    # Expose superevent
    cmd = 'expose {sid}'.format(sid=sid)
    try:
        response = run_CLI_test(cmd)
        out = response.json()

        # Check response-- this will be 200 for admins and
        # em_advocates, or 403 for everyone else, like th
        # gitlab runner.
        assert response.status_code == 200
        assert isinstance(out, list)
        assert len(out) == 3
    except HTTPError as e:
        assert e.status_code == 403

    # Hide superevent
    cmd = 'hide {sid}'.format(sid=sid)
    try:
        response = run_CLI_test(cmd)
        out = response.json()

        # Check response-- same rules apply for exposing:
        # Note that unauthorized users will get the 403 response
        # at the permissions check even if the superevent is not
        # exposed. So if the previous test 200'ed, then in theory
        # this one will too if everything worked.
        assert response.status_code == 200
        assert isinstance(out, list)
        assert len(out) == 0
    except HTTPError as e:
        assert e.status_code == 403


def test_voevents():
    """Test all things related to VOEvents with the CLI"""
    # Setup: create event
    cmd = 'create event Test gstlal {event_file} --offline' \
        .format(event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response = run_CLI_test(cmd)
    assert response.status_code == 201
    out = response.json()
    gid = out['graceid']

    # Setup: upload fake skymap
    filename = os.path.join(TEST_DATA_DIR, 'test_file.txt')
    cmd = 'create log {gid} "fake skymap" {filename}'.format(
        gid=gid, filename=filename)
    response = run_CLI_test(cmd)
    assert response.status_code == 201

    # Creation ----------------------------------------------------------------
    cmd = ('create voevent {gid} UP --skymap-filename={skymap} --hardware-inj '
           ' --skymap-type=update --prob-has-ns={pns} --bbh={bbh}').format(
        gid=gid, pns=0.1, bbh=0.3, skymap=os.path.basename(filename))
    response = run_CLI_test(cmd)
    out = response.json()
    voevent_N = out['N']

    # Check response
    assert response.status_code == 201
    assert out['N'] == voevent_N
    assert out['voevent_type'] == 'UP'

    # Get ---------------------------------------------------------------------
    cmd = 'get voevent {gid} {N}'.format(gid=gid, N=voevent_N)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check response
    assert response.status_code == 200
    assert out['N'] == voevent_N
    assert out['voevent_type'] == 'UP'

    # List --------------------------------------------------------------------
    cmd = 'list voevents {gid}'.format(gid=gid)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check response
    assert response.status_code == 200
    assert len(out['voevents']) == 1
    assert out['voevents'][0]['N'] == voevent_N
    assert out['voevents'][0]['voevent_type'] == 'UP'


def test_emobservations():
    """Test all things related to EM observations with the CLI"""
    # Setup: create event
    cmd = 'create event Test gstlal {event_file} --offline' \
        .format(event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response = run_CLI_test(cmd)
    assert response.status_code == 201
    out = response.json()
    gid = out['graceid']

    # Creation ----------------------------------------------------------------
    now = datetime.datetime.now()
    group = 'AGILE'
    ra_list = [1, 2, 3, 4]
    ra_width = 0.1
    dec_width = 0.2
    duration = 10
    dec_list = [5, 6, 7, 8]
    start_time_list = list(
        map(lambda i: (now + datetime.timedelta(seconds=i)).isoformat(),
            [0, 1, 2, 3])
    )

    # Compile command
    def join_s(s):
        return ",".join([str(i) for i in s]) if isinstance(s, list) else s

    cmd = ('create emobservation {gid} {group} {ra_list} {ra_width_list} '
           '{dec_list} {dec_width_list} {start_time_list} {duration_list}') \
        .format(gid=gid, group=group, ra_list=join_s(ra_list),
                ra_width_list=join_s(ra_width), dec_list=join_s(dec_list),
                dec_width_list=join_s(dec_width),
                duration_list=join_s(duration),
                start_time_list=join_s(start_time_list))
    cmd += " --comment='fake comment'"
    response = run_CLI_test(cmd)
    print("yay1", response)
    out = response.json()
    print("yay2", out)
    print("yay3", type(out))
    emo_N = out['N']

    # Check response
    assert response.status_code == 201
    assert out['comment'] == 'fake comment'
    assert len(out['footprints']) == len(ra_list)
    assert out['group'] == group

    # Get ---------------------------------------------------------------------
    cmd = 'get emobservation {gid} {N}'.format(gid=gid, N=emo_N)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check response
    assert response.status_code == 200
    assert out['N'] == emo_N
    assert out['comment'] == 'fake comment'
    assert len(out['footprints']) == len(ra_list)
    assert out['group'] == group

    # List --------------------------------------------------------------------
    cmd = 'list emobservations {gid}'.format(gid=gid)
    response = run_CLI_test(cmd)
    out = response.json()

    # Check response
    assert response.status_code == 200
    assert out['observations'][0]['N'] == emo_N
    assert out['observations'][0]['comment'] == 'fake comment'
    assert len(out['observations'][0]['footprints']) == len(ra_list)
    assert out['observations'][0]['group'] == group


def test_signoffs():
    """Test all things related to signoffs with the CLI"""
    # Setup: create event
    cmd = 'create event Test gstlal {event_file}'.format(
        event_file=os.path.join(TEST_DATA_DIR, 'cbc-lm.xml'))
    response = run_CLI_test(cmd)
    assert response.status_code == 201
    gid = response.json()['graceid']

    # Setup: create superevent
    cmd = ('create superevent 1 2 3 {p_ev} --labels=INJ,EM_READY '
           '--category=Test').format(p_ev=gid)
    response = run_CLI_test(cmd)
    out = response.json()
    assert response.status_code == 201
    sid = out['superevent_id']

    # Setup: apply ADVREQ label
    cmd = 'add label {sid} ADVREQ'.format(sid=sid)
    response = run_CLI_test(cmd)
    out = response.json()
    assert response.status_code == 201

    # Create signoff ----------------------------------------------------------
    cmd = 'create signoff {sid} ADV OK "looks good"'.format(sid=sid)
    try:
        response = run_CLI_test(cmd)
        out = response.json()

        # Check response
        assert response.status_code == 201
        assert out['status'] == 'OK'
        assert out['comment'] == 'looks good'
        assert out['instrument'] == ''
        assert out['signoff_type'] == 'ADV'

        # Define this session as an authorized session. The rest of the tests
        # are sensible only if this one passes. This is admittedly something
        # that needs to be run offline by an admin unless i can think of a more
        # clever way of doing it.
        authorized_user = True

    except HTTPError as e:
        assert e.status_code == 403
        authorized_user = False

    if authorized_user:
        # Get signoff ------------------------------------------------------
        cmd = 'get signoff {sid} ADV'.format(sid=sid)
        response = run_CLI_test(cmd)
        out = response.json()

        # Check response
        assert response.status_code == 200
        assert out['status'] == 'OK'
        assert out['comment'] == 'looks good'
        assert out['instrument'] == ''
        assert out['signoff_type'] == 'ADV'

        # List signoff -----------------------------------------------------
        cmd = 'list signoffs {sid}'.format(sid=sid)
        response = run_CLI_test(cmd)
        out = response.json()

        # Check response
        assert response.status_code == 200
        assert len(out['signoffs']) == 1
        assert out['signoffs'][0]['status'] == 'OK'
        assert out['signoffs'][0]['comment'] == 'looks good'
        assert out['signoffs'][0]['instrument'] == ''
        assert out['signoffs'][0]['signoff_type'] == 'ADV'

        # Update signoff ---------------------------------------------------
        cmd = 'update signoff {sid} ADV --status=NO --comment="new"'.format(
            sid=sid)
        response = run_CLI_test(cmd)
        out = response.json()

        # Check response
        assert response.status_code == 200
        assert out['status'] == 'NO'
        assert out['comment'] == 'new'
        assert out['instrument'] == ''
        assert out['signoff_type'] == 'ADV'

        # Delete signoff ---------------------------------------------------
        cmd = 'delete signoff {sid} ADV'.format(sid=sid)
        response = run_CLI_test(cmd)

        # Check response
        assert response.status_code == 204

        # List signoff again -----------------------------------------------
        cmd = 'list signoffs {sid}'.format(sid=sid)
        response = run_CLI_test(cmd)
        out = response.json()

        # Check response
        assert response.status_code == 200
        assert len(out['signoffs']) == 0


def test_ping():
    """Test ping with the CLI"""
    response = run_CLI_test('ping')
    assert response.startswith('Response from {0}'.format(TEST_SERVICE))
    assert response.endswith('200 OK')


def test_credentials():
    """Test getting credentials from server"""
    response = run_CLI_test('credentials server')
    assert response.status_code == 200


# Subcommands for testing 'info' command
CMDS = ['emgroups', 'groups', 'labels', 'pipelines', 'searches',
        'server_version', 'signoff_statuses', 'signoff_types',
        'superevent_categories', 'voevent_types']
@pytest.mark.parametrize("cmd", CMDS)  # noqa: E302
def test_info(cmd):
    """Test 'info' command with CLI"""
    response = run_CLI_test('info ' + cmd)

    # Can't really check content of response but we can make sure we got
    # something
    assert response != 'Data not found on server.'



# Commonly used argument parsers
import argparse
import textwrap

from .utils import parse_delimited_string, parse_delimited_string_or_single

# Custom parsers for specific arguments

# Object ID parser
object_id_parser = argparse.ArgumentParser(add_help=False)
object_id_parser.add_argument(
    "object_id", type=str, help="Event graceid or superevent id"
)

# Graceid parser
graceid_parser = argparse.ArgumentParser(add_help=False)
graceid_parser.add_argument("graceid", type=str, help="Event graceid")

# Superevent id parser
superevent_id_parser = argparse.ArgumentParser(add_help=False)
superevent_id_parser.add_argument(
    "superevent_id", type=str, help="Superevent id"
)

# Tag parser
tag_parser = argparse.ArgumentParser(add_help=False)
tag_parser.add_argument(
    "-t", "--tag-name", dest="tag_name",
    type=parse_delimited_string_or_single, default=None,
    help="Tag name or comma-separated list of tag names"
)
tag_parser.add_argument(
    "-d", "--tag-display-name", dest="tag_display_name",
    type=parse_delimited_string_or_single, default=None,
    help=("Tag display name or comma-separated list of tag display names "
          "(ignored for existing tags)")
)

# Label parser
label_parser = argparse.ArgumentParser(add_help=False)
label_parser.add_argument(
    "-l", "--labels", dest="labels",
    type=parse_delimited_string, default=None,
    help="Label name or comma-separated list of label names"
)

# Comment parser
comment_parser = argparse.ArgumentParser(add_help=False)
comment_parser.add_argument(
    "comment", type=str, default="",
    help="Comment to be added to the event or superevent log")


# Custom argument parser
class CustomHelpArgumentParser(argparse.ArgumentParser):
    """Must be paired with CommandBase or derived class"""
    _width = 80

    def __init__(self, subcommands=(), show_usage_args=True, *args, **kwargs):
        super(CustomHelpArgumentParser, self).__init__(*args, **kwargs)
        self.subcommands = subcommands
        self.show_usage_args = show_usage_args

    def format_subcommand_list(self):
        HEADER_INDENT = 2
        INITIAL_INDENT = 4
        CMD_COLUMN_WIDTH = 20

        # Header - write with hyphen separators initially so textwrap breaks
        # there and then replace with commas. Use double hyphens just to
        # be safe.
        subcommands = [sc for sc in self.subcommands if not sc[1].legacy]
        header = "Available commands:\n"
        header += textwrap.fill("{{{cmds}}}".format(cmds="--".join(
            [sc[0] for sc in subcommands])), width=self._width,
            initial_indent=' ' * HEADER_INDENT,
            subsequent_indent=' ' * (HEADER_INDENT + 1))
        header = header.replace('--', ',')

        # List of commands
        cmd_list = []
        ii_str = ' ' * INITIAL_INDENT + '{{cmd: <{ccw}}}'
        for cmd_name, cmd in subcommands:
            cmd_entry = ''
            if len(cmd_name) > CMD_COLUMN_WIDTH:
                # Handle case where command name is longer than the
                # allowed length for two nicely formatted columns. In this
                # situation, we put the description on the next line.
                cmd_entry += ' ' * INITIAL_INDENT + cmd_name + '\n'
                cmd_name = ''
            cmd_entry_kwargs = {
                'initial_indent': ii_str.format(ccw=CMD_COLUMN_WIDTH).format(
                    cmd=cmd_name),
                'width': self._width,
                'subsequent_indent': ' ' * (CMD_COLUMN_WIDTH + INITIAL_INDENT),
            }
            cmd_entry += textwrap.fill(cmd.description, **cmd_entry_kwargs)
            cmd_list.append(cmd_entry)

        return header + "\n\n" + "\n".join(cmd_list)

    def format_usage(self):
        # Run super
        usage = super(CustomHelpArgumentParser, self).format_usage()
        # Substitute [command] -> <command> and optionally add [<args>]
        usage = usage.replace('[command]', "<command>{args}".format(
            args=(" [<args>]" if self.show_usage_args else "")))
        return usage

    def format_help(self):
        formatter = self._get_formatter()

        # usage
        # CUSTOM: format usage
        formatter.add_usage(self.format_usage(), self._actions,
                            self._mutually_exclusive_groups, prefix='')

        # description
        # CUSTOM:
        formatter.add_text(textwrap.fill(self.description, width=self._width))

        # positionals, optionals, and user-defined groups
        # CUSTOM: skip positionals since they will only correspond
        # to the commands, which we will list ourselves.
        for action_group in self._action_groups:
            if action_group.title == 'positional arguments':
                continue
            formatter.start_section(action_group.title)
            formatter.add_text(action_group.description)
            formatter.add_arguments(action_group._group_actions)
            formatter.end_section()

        # CUSTOM: add list of subcommands
        if self.subcommands:
            formatter.add_text(self.format_subcommand_list())

        # epilog
        formatter.add_text(self.epilog)

        # determine help from format above
        return formatter.format_help()



import argparse
from six import with_metaclass
import sys

from . import command_registry
from ..parsers import CustomHelpArgumentParser


class CommandBase(object):
    """Class for a base or intermediate command"""
    name = None
    description = ""
    subcommands = []
    show_usage_args = True
    legacy = False

    def __init__(self, base_prog=None, parent_prog=None):
        # Get prog name
        prog = self.get_prog(parent_prog)

        # Set up list of tuples of subcommand name and the subcommand
        # object itself, sorted by name
        self.subcommand_list = sorted([(sc.name, sc) for sc in
                                      self.subcommands])

        # Set up dict of subcommands, where key is name and value is
        # the subcommand
        self.subcommand_dict = {sc.name: sc for sc in self.subcommands}

        # Set up parser description; add period to end if it doesn't have one
        description = getattr(self, 'long_description', self.description)
        if description and description[-1] != '.':
            description += '.'

        # Set up main parser
        parser_kwargs = {
            'description': description,
            'show_usage_args': self.show_usage_args,
            'subcommands': self.subcommand_list,
        }
        if prog:
            parser_kwargs['prog'] = prog
        parser = CustomHelpArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
            **parser_kwargs)
        parser = self.add_custom_arguments(parser)

        # Attach parser to instance
        self.parser = parser

        # Set up epilog
        parser.epilog = ("\n\nUse '{prog} <command> --help' to read about a "
                         "specific command.").format(prog=parser.prog)

        # Store base prog name
        self.base_prog = base_prog

    def __call__(self, client, args=None):
        # Parse args
        cmd, main_args, cmd_args = self.parse_args(args)
        return cmd(client, cmd_args)

    def parse_args(self, args):
        # Modify args to allow passing --help flag to subcommands
        modified_args = [arg for arg in args if (arg != '--help'
                         and arg != '-h')]

        # Check if help is requested (or needed)
        send_help = False
        if (len(modified_args) < len(args)):
            send_help = True

        # Parse args
        main_args, cmd_args = self.parser.parse_known_args(modified_args)

        # Store the args specific to this level
        self.args = main_args

        # If send_help and no command is specified, print help and exit
        if main_args.command is None:
            if send_help:
                self.parser.print_help()
                sys.exit(0)
            else:
                self.parser.error('too few arguments')

        # Get subcommand and instantiate it
        try:
            sc_index = [sc.name for sc in self.subcommands].index(
                main_args.command)
        except ValueError:
            # Command not found
            error_str = ("Command '{prog} {cmd}' not found. Do "
                         "'{prog} --help' to see available commands.\n") \
                .format(prog=self.parser.prog, cmd=main_args.command)
            print(error_str)
            sys.exit(1)

        # If base_prog is not set, assume this is the base prog
        base_prog = self.base_prog or self.parser.prog
        subcommand = self.subcommands[sc_index](parent_prog=self.parser.prog,
                                                base_prog=base_prog)

        # Append --help to cmd_args, if needed
        if send_help:
            cmd_args.append('--help')

        return subcommand, main_args, cmd_args

    def add_custom_arguments(self, parser):
        parser.add_argument("command", nargs='?')
        return parser

    def get_prog(self, parent_prog=None):
        if parent_prog:
            prog = "{parent} {name}".format(parent=parent_prog, name=self.name)
        elif self.name is None:
            return None
        else:
            prog = self.name
        return prog


class SubCommandBase(object):
    """Class for a subcommand"""
    parent_parsers = ()
    legacy = False

    def __init__(self, base_prog=None, parent_prog=None):
        # Store base prog name
        self.base_prog = base_prog

        # Description - if no long description is given, use the description.
        # Add a period if it doesn't end in one; this looks better in the
        # individual command description, but not in the list of commands
        description = getattr(self, 'long_description', self.description)
        if description and description[-1] != '.':
            description += '.'

        # Set up parser
        parser = argparse.ArgumentParser(
            description=description,
            parents=self.parent_parsers,
            prog=self.get_prog(parent_prog),
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
        parser = self.add_custom_arguments(parser)
        self.parser = parser

    def __call__(self, client, args):
        parsed_args = self.parser.parse_args(args)
        return self.run(client, parsed_args)

    def get_prog(self, parent_prog=None):
        if parent_prog:
            prog = "{parent} {name}".format(parent=parent_prog, name=self.name)
        elif self.name is None:
            return None
        else:
            prog = self.name
        return prog

    def add_custom_arguments(self, parser):
        return parser

    def run(self, client, args):
        return NotImplemented


###############################################################################
# Registering commands
###############################################################################
class RegisteredCommandMeta(type):
    def __new__(meta, name, bases, attrs):
        cls = type.__new__(meta, name, bases, attrs)
        if (hasattr(cls, '_register') and callable(cls._register)):
            if ('Base' not in name):
                cls._register()
        return cls


class RegisteredCommandBase(
    with_metaclass(RegisteredCommandMeta, CommandBase)
):

    @classmethod
    def _register(cls):
        if hasattr(cls, '_registry'):
            cls._registry.append(cls)
        else:
            command_registry.append(cls)


class RegisteredSubCommandBase(SubCommandBase, RegisteredCommandBase):
    pass



import re
import pytest
import six
try:
    from unittest import mock
except ImportError:
    import mock

# Add module-level mark
pytestmark = pytest.mark.cli


def test_cli_subcommands_content(expected_subcommands, CLI):
    """Ensure that subcommands for main CLI are what we expect"""

    # Check content
    subcommand_names = [sc.name for sc in CLI.subcommands if not sc.legacy]
    assert sorted(subcommand_names) == sorted(expected_subcommands)


def test_cli_bad_subcommand(CLI, capsys):
    """Test handling of bad subcommand input to CLI"""
    cmd = 'fake_subcommand'
    with pytest.raises(SystemExit) as excinfo:
        CLI([cmd])

    # Check exit code
    assert excinfo.value.code == 1

    # Get output
    stdout = capsys.readouterr().out
    assert "{cmd}' not found".format(cmd=cmd) in stdout


def test_cli_help(CLI):
    """Test CLI help functionality"""
    print_help_func_str = \
        'ligo.gracedb.cli.parsers.CustomHelpArgumentParser.print_help'
    with mock.patch(print_help_func_str) as mock_print_help:
        with pytest.raises(SystemExit) as excinfo:
            CLI(['--help'])

    # Make sure print_help is called once
    assert mock_print_help.call_count == 1

    # Check exit code
    assert excinfo.value.code == 0


def test_cli_help_content(CLI, expected_subcommands, capsys):
    """Check CLI help content"""
    with pytest.raises(SystemExit) as excinfo:
        CLI(['--help'])

    # Check exit code
    assert excinfo.value.code == 0

    # Get content of help message
    stdout = capsys.readouterr().out

    # Verify entry for each command
    cmd_entry_regex = r'\s+{cmd}\s+'
    available_commands = stdout[stdout.index('Available commands'):]
    for cmd in expected_subcommands:
        assert re.search(cmd_entry_regex.format(cmd=cmd), available_commands)


def test_cli_client_setup(CLI):
    """Test translation of CLI arguments into client instantiation"""
    arg_dict = {
        'service-url': 'testserver.com',
        'proxy': 'testproxy.com:123',
        'username': 'user',
        'password': 'passw0rd',
        'creds': 'cert_file,key_file',
    }
    args = ['--{k}={v}'.format(k=k, v=v) for k, v in six.iteritems(arg_dict)]

    client_class = 'ligo.gracedb.cli.client.CommandLineClient'
    with mock.patch(client_class) as mock_client:
        cli_args, cmd_args = CLI.parser.parse_known_args(args)
        CLI.set_up_client(cli_args)

    # Check call count
    assert mock_client.call_count == 1

    # Check args used to initialize client
    cli_args, cli_kwargs = mock_client.call_args
    assert cli_args == ()
    assert cli_kwargs['service_url'] == arg_dict['service-url']
    assert cli_kwargs['username'] == arg_dict['username']
    assert cli_kwargs['password'] == arg_dict['password']
    assert cli_kwargs['cred'] == arg_dict['creds'].split(',')



try:
    from unittest import mock
except ImportError:  # python < 3
    import mock

import pytest


@pytest.mark.parametrize("is_event", [True, False])
def test_get_tags(safe_client, is_event):
    if is_event:
        obj_id = 'T123456'
    else:
        obj_id = 'TS190302abc'
    N = 2

    # Set up templates mock
    mock_template = mock.MagicMock()
    if is_event:
        template_key = 'taglist-template'
    else:
        template_key = 'superevent-log-tag-list-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    # Set up allowed_labels mock
    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.tags(obj_id, N)

    # Test call args to get() method
    get_call_args, get_call_kwargs = mock_get.call_args
    assert len(get_call_args) == 1
    assert get_call_kwargs == {}

    # Test template call kwargs
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == 2
    assert template_call_kwargs['N'] == N
    if is_event:
        assert template_call_kwargs['graceid'] == obj_id
    else:
        assert template_call_kwargs['superevent_id'] == obj_id


@pytest.mark.parametrize(
    "is_event,display_name",
    [
        (True, None),
        (False, None),
        (True, 'test_disp'),
        (False, 'test_disp'),
    ]
)
def test_add_tag(safe_client, is_event, display_name):
    if is_event:
        obj_id = 'T123456'
        request_func = 'ligo.gracedb.client.GraceDBClient.put'
    else:
        obj_id = 'TS190302abc'
        request_func = 'ligo.gracedb.client.GraceDBClient.post'
    N = 7
    tag_name = 'test_tag'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if is_event:
        template_key = 'tag-template'
    else:
        template_key = 'superevent-log-tag-list-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    with mock.patch(request_func) as mock_rf, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.addTag(obj_id, N, tag_name, displayName=display_name)

    rf_call_args, rf_call_kwargs = mock_rf.call_args
    assert len(rf_call_args) == 1
    assert len(rf_call_kwargs) == 1
    assert 'data' in rf_call_kwargs
    if is_event:
        if display_name:
            assert len(rf_call_kwargs['data']) == 1
            assert rf_call_kwargs['data']['displayName'] == display_name
        else:
            assert rf_call_kwargs['data'] == {}
    else:
        assert rf_call_kwargs['data']['name'] == tag_name
        if display_name:
            assert len(rf_call_kwargs['data']) == 2
            assert rf_call_kwargs['data']['displayName'] == display_name
        else:
            assert len(rf_call_kwargs['data']) == 1

    # Test template call kwargs
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert template_call_kwargs['N'] == N
    if is_event:
        assert len(template_call_kwargs) == 3
        assert template_call_kwargs['graceid'] == obj_id
        assert template_call_kwargs['tag_name'] == tag_name
    else:
        assert len(template_call_kwargs) == 2
        assert template_call_kwargs['superevent_id'] == obj_id


@pytest.mark.parametrize("is_event", [True, False])
def test_remove_tag(safe_client, is_event):
    if is_event:
        obj_id = 'T123456'
    else:
        obj_id = 'TS190302abc'
    N = 7
    tag_name = 'test_tag'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if is_event:
        template_key = 'tag-template'
    else:
        template_key = 'superevent-log-tag-detail-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    # Set up allowed_labels mock
    with mock.patch('ligo.gracedb.rest.GraceDb.delete') as mock_delete, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.removeTag(obj_id, N, tag_name)

    delete_call_args, delete_call_kwargs = mock_delete.call_args
    assert len(delete_call_args) == 1
    assert len(delete_call_kwargs) == 0

    # Test template call kwargs
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == 3
    assert template_call_kwargs['N'] == N
    assert template_call_kwargs['tag_name'] == tag_name
    if is_event:
        assert template_call_kwargs['graceid'] == obj_id
    else:
        assert template_call_kwargs['superevent_id'] == obj_id



import os
import textwrap

from .base import RegisteredCommandBase, RegisteredSubCommandBase
from ..parsers import (
    object_id_parser, comment_parser, tag_parser,
    superevent_id_parser,
)
from ..utils import parse_delimited_string, parse_delimited_string_or_single

# Command registry - don't touch!
registry = []


###############################################################################
# Utilities
###############################################################################
def delimited_float(s):
    return parse_delimited_string(s, cast=float)


def delimited_or_single_float(s):
    return parse_delimited_string_or_single(s, cast=float)


###############################################################################
# Base command
###############################################################################
class CreateCommand(RegisteredCommandBase):
    name = "create"
    description = textwrap.dedent("""\
        Create an event, superevent, log entry, signoff, EM observation,
        or VOEvent
    """).rstrip()
    subcommands = registry


###############################################################################
# Subcommands - registered to base command automatically
###############################################################################
class CreateChildBase(RegisteredSubCommandBase):
    _registry = registry


class CreateEmobservationCommand(CreateChildBase):
    name = "emobservation"
    description = "Upload EM observation data for an event or superevent"
    parent_parsers = (object_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "group",
            type=str,
            help=("Name of EM MOU group making the observation "
                  "(do '{prog} info emgroups' to see options)")
            .format(prog=self.base_prog)
        )
        parser.add_argument(
            "right_ascension",
            type=delimited_float,
            help=("Comma-separated list of right ascension coordinates "
                  "(degrees)")
        )
        parser.add_argument(
            "right_ascension_width",
            type=delimited_or_single_float,
            help=("Comma-separated list of right ascension measurement widths "
                  "OR a single number if all measurements have the same width "
                  "(degrees)")
        )
        parser.add_argument(
            "declination",
            type=delimited_float,
            help=("Comma-separated list of declination coordinates (degrees)")
        )
        parser.add_argument(
            "declination_width",
            type=delimited_or_single_float,
            help=("Comma-separated list of declination measurement widths "
                  "OR a single number if all measurements have the same width "
                  "(degrees)")
        )
        parser.add_argument(
            "start_time",
            type=parse_delimited_string,
            help=("Comma-separated list of measurement start times in ISO "
                  "8601 format")
        )
        parser.add_argument(
            "duration",
            type=delimited_or_single_float,
            help=("List of exposure times OR a single number if all "
                  "measurements have the same exposure (seconds)")
        )
        parser.add_argument(
            "--comment", type=str, help="Comment about the observation"
        )
        return parser

    def run(self, client, args):
        return client.writeEMObservation(
            args.object_id, args.group, args.right_ascension,
            args.right_ascension_width, args.declination,
            args.declination_width, args.start_time, args.duration,
            comment=args.comment
        )


class CreateEventCommand(CreateChildBase):
    name = "event"
    description = "Create an event"
    long_description = textwrap.dedent("""\
        Create an event on the server by uploading an event data file.
    """).rstrip()

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "group",
            type=str,
            help=("Analysis group which identified the event "
                  "(do '{prog} info groups' to see options)")
            .format(prog=self.base_prog)
        )
        parser.add_argument(
            "pipeline",
            type=str,
            help=("Analysis pipeline which identified the event "
                  "(do '{prog} info pipelines' to see options)")
            .format(prog=self.base_prog)
        )
        parser.add_argument("event_file", type=str, help="Event data file")
        parser.add_argument(
            "search",
            type=str,
            nargs='?',
            help=("Search type (do '{prog} info searches' to see "
                  "options)").format(prog=self.base_prog)
        )
        parser.add_argument(
            "--labels",
            type=parse_delimited_string_or_single,
            help=("Label or comma-separated list of labels to apply to the "
                  "event upon creation")
        )
        parser.add_argument(
            "--offline",
            action="store_true",
            default=False,
            help="Signifies that the event was found by an offline analysis"
        )
        return parser

    def run(self, client, args):
        # Handle case where user reverses order of search and event file
        # We don't print a warning (for now) since this is mimicing
        # legacy behavior
        event_file = args.event_file
        search = args.search
        if event_file in client.searches and os.path.isfile(search):
            temp = event_file
            event_file = search
            search = temp

        return client.createEvent(
            args.group, args.pipeline, event_file, search=search,
            offline=args.offline, labels=args.labels
        )


class CreateLogCommand(CreateChildBase):
    name = "log"
    description = "Create a log entry, with optional file upload"
    long_description = textwrap.dedent("""\
        Annotate an event or superevent by creating a log entry. The entry
        may include data, links, or other comments. Files may be uploaded
        along with the log entry. Tags can be applied to the log entry
        upon its creation.
    """).rstrip()
    parent_parsers = (object_id_parser, comment_parser, tag_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument("filename", type=str, nargs='?',
                            help="Path to file to be uploaded (optional)")
        return parser

    def run(self, client, args):
        return client.writeLog(
            args.object_id, args.comment, filename=args.filename,
            tag_name=args.tag_name, displayName=args.tag_display_name
        )


class CreateSignoffCommand(CreateChildBase):
    name = "signoff"
    description = "Create an operator or advocate signoff"
    long_description = textwrap.dedent("""\
        Create an operator or advocate signoff. Only allowed for superevents
        which are labeled with the corresponding *OPS or ADVREQ labels. Event
        signoff creation is not presently implemented.
    """).rstrip()
    parent_parsers = (superevent_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "signoff_type",
            type=str,
            help=("Signoff type (do '{prog} info signoff_types') to see "
                  "options").format(prog=self.base_prog)
        )
        parser.add_argument(
            "signoff_status",
            type=str,
            help=("Signoff type (do '{prog} info signoff_statuses') to see "
                  "options").format(prog=self.base_prog)
        )
        parser.add_argument(
            "comment",
            type=str,
            help="Justification for signoff status"
        )
        parser.add_argument(
            "instrument",
            type=str,
            nargs='?',
            help=("Instrument code (do '{prog} info instruments' to see "
                  "options). Required for operator signoffs.")
            .format(prog=self.base_prog)
        )
        return parser

    def run(self, client, args):
        instrument = args.instrument or ''  # Convert None to ''
        return client.create_signoff(
            args.superevent_id, args.signoff_type, args.signoff_status,
            args.comment, instrument=instrument
        )


class CreateSupereventCommand(CreateChildBase):
    name = "superevent"
    description = "Create a superevent"
    long_description = textwrap.dedent("""\
        Create a superevent by specifying t_start, t_0, t_end, and the
        graceid of an event to set as the 'preferred_event'. An additional
        list of event graceids to include in the superevent can also be
        supplied.
    """).rstrip()

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "t_start",
            type=float,
            help="t_start of superevent"
        )
        parser.add_argument("t_0", type=float, help="t_0 of superevent")
        parser.add_argument("t_end", type=float, help="t_end of superevent")
        parser.add_argument(
            "preferred_event",
            type=str,
            help="Graceid of the preferred event"
        )
        parser.add_argument(
            "--category",
            type=str,
            default="production",
            help=("Superevent category (do '{prog} info superevent_categories "
                  "to see options'").format(prog=self.base_prog)
        )
        parser.add_argument(
            "--events",
            type=parse_delimited_string_or_single,
            help=("Comma-separated list of graceids corresponding to events "
                  "which should be added to this superevent")
        )
        parser.add_argument(
            "--labels",
            type=parse_delimited_string_or_single,
            help=("Label or comma-separated list of labels to apply to the "
                  "superevent upon creation")
        )
        return parser

    def run(self, client, args):
        return client.createSuperevent(
            args.t_start, args.t_0, args.t_end, args.preferred_event,
            category=args.category, events=args.events, labels=args.labels
        )


class CreateVoeventCommand(CreateChildBase):
    name = "voevent"
    description = "Create a VOEvent for an event or superevent"
    parent_parsers = (object_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "voevent_type",
            type=str,
            help=("VOEvent type (do '{prog} info voevent_types to"
                  "see options'").format(prog=self.base_prog)
        )
        parser.add_argument(
            "--skymap-type",
            type=str,
            help="Skymap type (required for VOEvents which include a skymap)"
        )
        parser.add_argument(
            "--skymap-filename",
            type=str,
            help=("Name of skymap file on the server (required for initial "
                  "and update alerts, optional for preliminary)")
        )
        parser.add_argument(
            "--external",
            action='store_true',
            default=False,
            help=("Signifies that the VOEvent should be distributed outside "
                  "the LVC")
        )
        parser.add_argument(
            "--open-alert",
            action='store_true',
            default=False,
            help="Signifies that the candidate is an open alert"
        )
        parser.add_argument(
            "--hardware-inj",
            action='store_true',
            default=False,
            help="The candidate is a hardware injection"
        )
        parser.add_argument(
            "--coinc-comment",
            action='store_true',
            default=False,
            help="The candidate has a possible counterpart GRB"
        )
        parser.add_argument(
            "--prob-has-ns",
            type=float,
            default=None,
            help=("Probability that one object in the binary has mass less "
                  "than 3 M_sun (0.0 - 1.0)")
        )
        parser.add_argument(
            "--prob-has-remnant",
            type=float,
            default=None,
            help=("Probability that there is matter in the surroundings of "
                  "the central object (0.0 - 1.0)")
        )
        parser.add_argument(
            "--bns",
            type=float,
            default=None,
            help=("Probability that the source is a binary neutron star "
                  "merger (0.0 - 1.0)")
        )
        parser.add_argument(
            "--nsbh",
            type=float,
            default=None,
            help=("Probability that the source is a neutron star-black hole "
                  "merger (0.0 - 1.0)")
        )
        parser.add_argument(
            "--bbh",
            type=float,
            default=None,
            help=("Probability that the source is a binary black hole "
                  "merger (0.0 - 1.0)")
        )
        parser.add_argument(
            "--terrestrial",
            type=float,
            default=None,
            help="Probability that the source is terrestrial (0.0 - 1.0)"
        )
        parser.add_argument(
            "--mass-gap",
            type=float,
            default=None,
            help=("Probability that at least one object in the binary is "
                  "between 3 and 5 solar masses")
        )
        return parser

    def run(self, client, args):
        return client.createVOEvent(
            args.object_id, args.voevent_type, skymap_type=args.skymap_type,
            skymap_filename=args.skymap_filename, internal=(not args.external),
            open_alert=args.open_alert, hardware_inj=args.hardware_inj,
            CoincComment=args.coinc_comment, ProbHasNS=args.prob_has_ns,
            ProbHasRemnant=args.prob_has_remnant, BNS=args.bns, NSBH=args.nsbh,
            BBH=args.bbh, Terrestrial=args.terrestrial, MassGap=args.mass_gap
        )



try:
    from unittest import mock
except ImportError:  # python < 3
    import mock

import pytest


@pytest.mark.parametrize(
    "log_number",
    ['str', 1.23, [1], (2,), "", {'key': 'value'}]
)
def test_bad_log_number(safe_client, log_number):
    err_str = 'log_number should be an int'
    with pytest.raises(TypeError, match=err_str):
        safe_client.logs('TS121212abc', log_number=log_number)


@pytest.mark.parametrize("log_number", [None, 12])
def test_superevent_logs(safe_client, log_number):
    superevent_id = 'TS190302abc'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if log_number:
        template_key = 'superevent-log-detail-template'
    else:
        template_key = 'superevent-log-list-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.logs(superevent_id, log_number=log_number)

    get_call_args, get_call_kwargs = mock_get.call_args
    assert len(get_call_args) == 1
    assert get_call_kwargs == {}

    # Test template call kwargs
    num_kwargs = 1
    if log_number:
        num_kwargs += 1
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == num_kwargs
    assert template_call_kwargs['superevent_id'] == superevent_id
    if log_number:
        assert template_call_kwargs['N'] == log_number


@pytest.mark.parametrize("log_number", [None, 12])
def test_event_logs(safe_client, log_number):
    graceid = 'T123456'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if log_number:
        template_key = 'event-log-detail-template'
    else:
        template_key = 'event-log-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.logs(graceid, log_number=log_number)

    get_call_args, get_call_kwargs = mock_get.call_args
    assert len(get_call_args) == 1
    assert get_call_kwargs == {}

    # Test template call kwargs
    num_kwargs = 1
    if log_number:
        num_kwargs += 1
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == num_kwargs
    assert template_call_kwargs['graceid'] == graceid
    if log_number:
        assert template_call_kwargs['N'] == log_number


@pytest.mark.parametrize(
    "is_event,obj_id,tag_name,display_name",
    [
        (True, "T123456", None, None),
        (False, "TS121212abc", None, None),
        (True, "T123456", 'test', None),
        (True, "T123456", ('new_tag',), None),
        (True, "T123456", ['tag1', 'tag2'], None),
        (False, "TS121212abc", 'tag1', 'tag1_disp'),
        (False, "TS121212abc", ['tag1'], 'tag1_disp'),
        (False, "TS121212abc", set(['tag1']), set(['tag1_disp'])),
    ]
)
def test_write_log(safe_client, is_event, obj_id, tag_name, display_name):
    # Comment
    comment = 'test'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if is_event:
        template_key = 'event-log-template'
    else:
        template_key = 'superevent-log-list-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.post') as mock_post, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.writeLog(obj_id, comment, tag_name=tag_name,
                             displayName=display_name)

    # Test call to self.post
    post_call_args, post_call_kwargs = mock_post.call_args
    assert len(post_call_args) == 1
    assert len(post_call_kwargs) == 2
    assert 'data' in post_call_kwargs
    assert 'files' in post_call_kwargs
    request_body = post_call_kwargs['data']
    assert request_body['comment'] == comment
    assert isinstance(request_body['tagname'], list)
    assert isinstance(request_body['displayName'], list)
    if not tag_name:
        assert request_body['tagname'] == []
    elif isinstance(tag_name, str):
        assert request_body['tagname'] == [tag_name]
    elif isinstance(tag_name, (list, tuple, set)):
        assert request_body['tagname'] == list(tag_name)
    if not display_name:
        assert request_body['displayName'] == []
    elif isinstance(display_name, str):
        assert request_body['displayName'] == [display_name]
    elif isinstance(tag_name, (list, tuple, set)):
        assert request_body['displayName'] == list(display_name)

    # Test template call kwargs
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == 1
    if is_event:
        call_key = 'graceid'
    else:
        call_key = 'superevent_id'
    assert template_call_kwargs[call_key] == obj_id


@pytest.mark.parametrize(
    "tag_name,legacy_tagname",
    [
        (None, None),
        (None, ['test_tag']),
        (['tag1'], None),
        (['tag1'], ['tag2']),
    ]
)
def test_write_log_legacy_tagname(safe_client, tag_name, legacy_tagname):
    # Comment
    comment = 'test'

    # Set up kwargs
    method_kwargs = {}
    if tag_name:
        method_kwargs['tag_name'] = tag_name
    if legacy_tagname:
        method_kwargs['tagname'] = legacy_tagname

    # Set up templates mock
    mock_template = mock.MagicMock()
    mock_template_dict = {'superevent-log-list-template': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.post') as mock_post, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.writeLog('TS121212abc', comment, **method_kwargs)

    # Test call to self.post
    post_call_args, post_call_kwargs = mock_post.call_args
    assert len(post_call_args) == 1
    assert len(post_call_kwargs) == 2
    assert 'data' in post_call_kwargs
    assert 'files' in post_call_kwargs
    request_body = post_call_kwargs['data']
    assert request_body['comment'] == comment
    if legacy_tagname and not tag_name:
        assert request_body['tagname'] == legacy_tagname
    else:
        if tag_name is None:
            tag_name = []
        assert request_body['tagname'] == tag_name


@pytest.mark.parametrize(
    "tag_name,display_name",
    [
        (None, 'tag1_disp'),
        ('tag1', ['tag1_disp1', 'tag1_disp2']),
        (['tag1', 'tag2'], 'tag1_disp'),
        (['tag1', 'tag2'], ['tag1_disp1', 'tag1_disp2', 'tag2_disp']),
    ]
)
def test_write_log_bad_tags(safe_client, tag_name, display_name):
    err_msg = ("For a list of tags, either provide no display names or a "
               "display name for each tag")
    with pytest.raises(ValueError, match=err_msg):
        safe_client.writeLog('TS121212abc', 'test', tag_name=tag_name,
                             displayName=display_name)



import logging
try:
    from unittest import mock
except ImportError:  # py < 3
    import mock
import time
import types

from ligo.gracedb.logging import GraceDbLogHandler


def test_logger(safe_client):
    # Data
    obj_id = 'T123456'
    comment = 'test comment'
    logger_name = 'testing'
    log_method = 'warning'

    # Set up logger
    logging.basicConfig()
    log = logging.getLogger(logger_name)
    log.propagate = False  # Don't write to console

    # For some reason, mocking with mock.patch() and a context manager
    # was not working well, so we're doing this instead.
    mock_write_log = mock.MagicMock()
    safe_client.writeLog = types.MethodType(mock_write_log, safe_client)

    # Set up the handler and use the logger
    handler = GraceDbLogHandler(safe_client, obj_id)
    log.addHandler(handler)
    getattr(log, log_method)(comment)

    # Check results - there must be something weird with threading going on
    # since if we don't sleep here, mock_write_log.call_args returns as None
    time.sleep(0.01)
    call_args, call_kwargs = mock_write_log.call_args
    assert mock_write_log.call_count == 1
    assert call_kwargs == {}
    assert len(call_args) == 3
    assert call_args[1] == obj_id
    expected_comment = '{lvl}:{logger_name}:{comment}'.format(
        lvl=log_method.upper(), logger_name=logger_name, comment=comment)
    assert expected_comment in call_args[2]



# -*- coding: utf-8 -*-
# Copyright (C) Brian Moe, Branson Stephens (2015)
#
# This file is part of gracedb
#
# gracedb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gracedb.  If not, see <http://www.gnu.org/licenses/>.
from io import open
# import mimetypes
import os
import six
import sys
from six.moves import map
from six.moves.urllib.parse import urlencode

from .exceptions import HTTPError
from .utils import event_or_superevent, dict_to_form_encoded, get_mimetype
from .client import GraceDBClient

DEFAULT_SERVICE_URL = "https://gracedb.ligo.org/api/"


# -----------------------------------------------------------------
# GraceDb REST client
# -----------------------------------------------------------------
class GraceDb(GraceDBClient):
    """GraceDb REST client class.

    Provides a client object for accessing the GraceDB server API.
    Various methods are provided for retrieving information about different
    objects and uploading information.

    Lookup of user credentials is done in the following order:

    #. If provided, import X.509 credentials from the certificate-key \
        pair or combined proxy file provided in the ``cred`` keyword arg.
    #. If provided, use the username and password provided in the \
        keyword arguments.
    #. If the ``X509_USER_CERT`` and ``X509_USER_KEY`` environment variables \
        are set, load the corresponding certificate and key.
    #. If the ``X509_USER_PROXY`` environment variable is set, load the \
        corresponding proxy file.
    #. Look for a X.509 proxy from ligo-proxy-init in the default location \
        (``/tmp/x509up_u${UID}``).
    #. Look for a certificate and key file in ``$HOME/.globus/usercert.pem`` \
        and ``$HOME/.globus/userkey.pem``.
    #. Look for a username and password for the server in ``$HOME/.netrc``.
    #. Continue with no authentication credentials.

    Args:
        url (:obj:`str`, optional): URL of server API root.
        cred (:obj:`tuple` or :obj:`str`, optional): a tuple or list of
            (``/path/to/cert/file``, ``/path/to/key/file``) or a single path to
            a combined proxy file. Used for X.509 authentication only.
        username (:obj:`str`, optional): username for basic auth.
        password (:obj:`str`, optional): password for basic auth.
        force_noauth (:obj:`bool`, optional): set to True if you want to
            skip credential lookup and use this client without
            authenticating to the server.
        fail_if_noauth (:obj:`bool`, optional): set to ``True`` if you want the
            client constructor to fail if no authentication credentials are
            provided or found.
        api_version (:obj:`str`, optional): choose the version of the server
            API to use.  At present, there is only one version, but this
            argument is provided with the expectation that this will change
            in the future.
        reload_certificate (:obj:`bool`, optional): if True, your certificate
            will be checked before each request whether it is within
            ``reload_buffer`` seconds of expiration, and if so, it will be
            reloaded. Useful for processes which may live longer than the
            certificate lifetime and have an automated method for certificate
            renewal. The path to the new/renewed certificate **must** be the
            same as for the old certificate.
        reload_buffer (:obj:`int`, optional): buffer (in seconds) for reloading
            a certificate in advance of its expiration. Only used if
            ``reload_certificate`` is ``True``.

    Examples:
        Instantiate a client to use the production GraceDB server:

        >>> g = GraceDb()

        /se another GraceDB server:

        >>> g = GraceDb(service_url='https://gracedb-playground.ligo.org/api/')

        Use a certificate and key in the non-default location:

        >>> g = GraceDb(cred=('/path/to/cert/file', '/path/to/key/file'))
    """
    def __init__(self, service_url=DEFAULT_SERVICE_URL,
                 cred=None, username=None, password=None,
                 force_noauth=False, fail_if_noauth=False, api_version=None,
                 reload_certificate=False, reload_buffer=300):
        """Create a client instance."""
        super(GraceDb, self).__init__(
            service_url,
            cred=cred, username=username, password=password,
            force_noauth=force_noauth, fail_if_noauth=fail_if_noauth,
            reload_certificate=reload_certificate, reload_buffer=reload_buffer
        )

        # Check version type
        if (api_version is not None and not
            isinstance(api_version, six.string_types)):
            # Raise error is not a string
            raise TypeError('api_version should be a string')

        # Sets default and versioned service URLs
        # (self._service_url, self._versioned_service_url)
        self._set_service_url(service_url, api_version)

        # Set version
        self._api_version = api_version

        # Set service_info to None, will be obtained from the server when
        # the user takes an action which needs this information.
        self._service_info = None

    def _set_service_url(self, service_url, api_version):
        """Sets versioned and unversioned service URLs"""
        # Make sure path ends with '/'
        if not service_url.endswith('/'):
            service_url += '/'

        # Default service url (unversioned)
        self._service_url = service_url

        # Versioned service url (if version provided)
        self._versioned_service_url = service_url
        if api_version and api_version != 'default':
            # If api_version is 'default', that is equivalent to not setting
            # the version and indicates that the user wants to use the
            # default/non-versioned API
            self._versioned_service_url += (api_version + '/')

    @property
    def service_url(self):
        # Will be removed in the future
        print("DEPRECATED: this attribute has been moved to '_service_url'")
        return self._service_url

    @property
    def service_info(self):
        """Gets the root API information."""
        if not self._service_info:
            # try-except block takes user-specified API version to use and
            # checks whether that version is available on the server
            try:
                r = self.request("GET", self._versioned_service_url)
            except HTTPError as e:
                # If we get a 404 error, that means that the versioned
                # service URL was not found. We assume that this happened
                # because the user requested an unavailable API version.
                if (e.status == 404):
                    # Get versions from unversioned API root
                    r = self.request("GET", self._service_url)
                    available_api_versions = r.json().get('API_VERSIONS', None)
                    if available_api_versions:
                        err_msg = ('Bad API version. Available versions for '
                                   'this server are: {0}').format(
                            available_api_versions)
                    else:
                        # Case where server doesn't have versions, for some
                        # reason.
                        err_msg = ('This server does not have a versioned '
                                   'API. Reinstantiate your client without a '
                                   'version.')

                    # Raise error
                    raise ValueError(err_msg)
                else:
                    # Not a 404 error, must be something else
                    raise e
            else:
                if r.status_code != 200:
                    raise HTTPError(r)
            self._service_info = r.json()
        return self._service_info

    @property
    def api_versions(self):
        """List of available API versions on the server."""
        return self.service_info.get('api-versions')

    @property
    def server_version(self):
        """Get the code version being run on the GraceDB server."""
        return self.service_info.get('server-version')

    @property
    def links(self):
        return self.service_info.get('links')

    @property
    def templates(self):
        return self.service_info.get('templates')

    @property
    def groups(self):
        """List of available analysis groups on the server."""
        return self.service_info.get('groups')

    @property
    def pipelines(self):
        """List of  available analysis pipelines on the server."""
        return self.service_info.get('pipelines')

    @property
    def searches(self):
        """List of available search types on the server."""
        return self.service_info.get('searches')

    # Would like to call this 'labels' to keep in line with how
    # other properties are named, but it's already used for a function.
    @property
    def allowed_labels(self):
        """List of available labels on the server."""
        return self.service_info.get('labels')

    @property
    def em_groups(self):
        """List of available EM groups on the server."""
        return self.service_info.get('em-groups')

    @property
    def voevent_types(self):
        """List of available VOEvent types on the server."""
        return self.service_info.get('voevent-types')

    @property
    def superevent_categories(self):
        """List of available superevent categories on the server."""
        return self.service_info.get('superevent-categories')

    @property
    def instruments(self):
        """List of available instruments on the server."""
        return self.service_info.get('instruments')

    @property
    def signoff_types(self):
        """List of available signoff types on the server."""
        return self.service_info.get('signoff-types')

    @property
    def signoff_statuses(self):
        """List of available signoff statuses on the server."""
        return self.service_info.get('signoff-statuses')

#   def request(self, method, url, body=None, headers=None, priming_url=None):
#       if (method.upper() in ['POST', 'PUT'] and self.auth_type == 'x509'):
#           priming_url = self._service_url
#       return super(GraceDb, self).request(
#           method, url, body, headers, priming_url)

    def _getCode(self, input_value, code_dict):
        """
        Check if input is valid and return coded version if it is
        code_dict is dict of {code: descriptive_name}
        """
        # Quick check for simple case where it's already coded
        if input_value in code_dict:
            return input_value

        # Loop over code_dict items, if we match either the key or
        # value (case-insensitive), return the code.
        input_lower = input_value.lower()
        for code, display in six.iteritems(code_dict):
            if (input_lower == code.lower()
                or input_lower == display.lower()):
                return code

        # Not found, return None
        return None

    # Search and filecontents are optional when creating an event.
    def createEvent(self, group, pipeline, filename, search=None, labels=None,
                    offline=False, filecontents=None, **kwargs):
        """Create a new event on the server.

        All LIGO-Virgo users can create events in the 'Test' group. Special
        permissions are required to create non-test events.

        Args:
            group (str): name of the analysis group which identified the
                candidate.
            pipeline (str): name of the analysis pipeline which identified the
                candidate.
            filename (str): path to event file to be uploaded. Use ``'-'`` to
                read from stdin.
            search (:obj:`str`, optional): type of search being run by the
                analysis pipeline.
            labels (:obj:`str` or :obj:`list[str]`, optional): label(s) to
                attach to the event upon creation. Should be a string (single
                label) or list of strings (multiple labels).
            offline (:obj:`bool`, optional): if ``True``, indicates that the
                event was found by an offline analysis.
            filecontents(:obj:`str`, optional): string to be uploaded to the
                server and saved into a file. If event data is uploaded via
                this mechanism, the ``filename`` argument is only used to
                set the name of the file once it is saved on the server.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = g.createEvent('CBC', 'gstlal', '/path/to/something.xml',
            ... labels='INJ', search='LowMass')
            >>> r.status_code
            201
        """
        if group not in self.groups:
            raise ValueError('bad group')
        if pipeline not in self.pipelines:
            raise ValueError('bad pipeline')
        if search and search not in self.searches:
            raise ValueError('bad search')
        # Process offline arg
        if not isinstance(offline, bool):
            raise TypeError('offline parameter should be a bool')
        # Process label args - convert non-empty strings to list
        # to ensure consistent processing
        if labels is not None:
            if isinstance(labels, six.string_types):
                # Convert to list
                labels = [labels]
            elif isinstance(labels, list):
                pass
            else:
                raise TypeError("labels arg is {0}, should be str or list"
                                .format(type(labels)))
            # Check labels against those in database
            for l in labels:
                if l not in self.allowed_labels:
                    raise ValueError(("Label '{0}' does not exist in the "
                                     "database").format(l))
        if filecontents is None:
            if filename == '-':
                filename = 'initial.data'
                filecontents = sys.stdin.read()
            else:
                filecontents = open(filename, 'rb')

        # Assemble dictionary before converting to multipart form
        # booleans can't be encoded, but the server will convert the
        # string.
        fields = {
            'group': group,
            'pipeline': pipeline,
            'offline': str(offline),
        }
        if search:
            fields.update({'search': search})
        if labels:
            fields.update({'labels': labels})

        # Update fields with additional keyword arguments
        for key, value in six.iteritems(kwargs):
            fields.update({key: value})

        # Assemble files dict and encode.
        files = {'eventFile': (filename,
                               filecontents,
                               get_mimetype(filename))}

        fields.update(files)
        fields = dict_to_form_encoded(fields)
        # fields = MultipartEncoder(fields=fields)
        # headers = {'Content-Type': fields.content_type}

        # Python httplib bug?  unicode link
        uri = str(self.links['events'])

        return self.post(uri, data=fields, files=files)

    def replaceEvent(self, graceid, filename, filecontents=None):
        """Replace an existing event by uploading a new event file.

        The event's parameters are updated from the new file. Only the user
        who originally created the event can update it.

        Args:
            graceid (str): GraceDB ID of the existing event
            filename (str): path to new event file
            filecontents(:obj:`str`, optional): string to be uploaded to the
                server and saved into a file. If event data is uploaded via
                this mechanism, the ``filename`` argument is only used to
                set the name of the file once it is saved on the server.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = g.replaceEvent('T101383', '/path/to/new/something.xml')
        """
        if filecontents is None:
            # Note: not allowing filename '-' here.  We want the event datafile
            # to be versioned.
            filecontents = open(filename, 'rb')
        return self.put(
            self.templates['event-detail-template'].format(graceid=graceid),
            files={'eventFile': (filename,
                                 filecontents,
                                 get_mimetype(filename))}
        )

    def update_grbevent(self, graceid, ra=None, dec=None, error_radius=None,
                        t90=None, redshift=None, designation=None):
        """Update a GRB event's parameters.

        This method only works on GRB events; i.e., External or Test events
        whose search is specified as 'GRB'.

        Only LIGO/Virgo users with permission to update GRB events will be
        able to utilize this method.

        Args:
            graceid (str): GraceDB ID of the existing event
            ra (:obj:`float`, optional): right ascension (degrees)
            dec (:obj:`float`, optional): declination (degrees)
            error_radius (:obj:`float`, optional): uncertainty in position
                as statistical ~1-sigma error radius from the reported right
                ascension and declination (degrees)
            t90 (:obj:`float`, optional): duration of the event in which 90%
                of the gamma-ray burst fluence was accumulated (seconds)
            redshift (:obj:`float`, optional): redshift
            designation (:obj:`str`, optional): name of the event, typically
                reported in GRByymmddx (where x is null, 'A', or 'B') or
                GRByymmddfff format, where fff is the three-digit fraction
                of the day

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = g.update_grbevent('E274955', redshift=2.33, ra=135.91)
            >>> r.status
            200
        """  # noqa: W605
        # Make sure that at least one parameter is provided
        if not (ra or dec or error_radius or t90 or redshift or designation):
            raise ValueError('Provide at least one of ra, dec, error_radius, '
                             't90, redshift, or designation')

        request_body = {}
        if ra is not None:
            request_body['ra'] = ra
        if dec is not None:
            request_body['dec'] = dec
        if error_radius is not None:
            request_body['error_radius'] = error_radius
        if t90 is not None:
            request_body['t90'] = t90
        if redshift is not None:
            request_body['redshift'] = redshift
        if designation is not None:
            request_body['designation'] = designation
        template = self.templates['update-grbevent-template']
        uri = template.format(graceid=graceid)
        return self.patch(uri, data=request_body)

    def event(self, graceid):
        """Get information about an individual event.

        Args:
            graceid (str): GraceDB ID of the event

        Returns:
            :class:`httplib.HTTPResponse`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> event_dict = g.event('T101383').json()
        """
        return self.get(
            self.templates['event-detail-template'].format(graceid=graceid)
        )

    def events(self, query=None, orderby=None, max_results=None, **kwargs):
        """Search for events which match a query.

        Information on forming queries is available here:
        https://gracedb.ligo.org/documentation/queries.html

        Args:
            query (:obj:`str`, optional): query string.
            orderby (:obj:`str`, optional): field to order the results by.
            max_results (:obj:`int`, optional): maximum number of results to
                return (default: all).

        Returns:
            :obj:`Iterator[dict]`

            An iterator which yields individual event dictionaries.

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> for event in g.events('ER5 submitter: "gstlalcbc"'):
            ...     print(event['graceid'], event['far'], event['gpstime'])
        """
        columns = kwargs.pop('columns', None)
        count = kwargs.pop('count', None)
        uri = self.links['events']
        qdict = {}
        if query:
            qdict['query'] = query
        if count is not None:
            qdict['count'] = count
        if orderby:
            qdict['sort'] = orderby
        if columns:
            qdict['columns'] = columns
        if qdict:
            uri += "?" + urlencode(qdict)
        n = 0
        while uri:
            response = self.get(uri).json()
            events = response.get('events', [])
            uri = response.get('links', {}).get('next')

            for event in events:
                n += 1
                if (max_results is not None and n > max_results):
                    break
                yield event

    def numEvents(self, query=None):
        """Get the number of events satisfying a query.

        Args:
            query (:obj:`str`, optional): query string.

        Returns:
            int: The number of events which matched the given query.

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> g.numEvents('ER5 submitter: "gstlalcbc"')
            213
        """
        uri = self.links['events']
        if query:
            uri += "?" + urlencode({'query': query})
        return self.get(uri).json()['numRows']

    def createSuperevent(self, t_start, t_0, t_end, preferred_event,
                         category='production', events=None, labels=None):
        r"""Create a superevent.

        All LIGO-Virgo users can create test superevents, but special
        permissions are required to create production and MDC superevents.

        Args:
            t_start (float): t\ :sub:`start` of the superevent
            t_0 (float): t\ :sub:`0` of the superevent
            t_end (float): t\ :sub:`end` of the superevent
            preferred_event (str): graceid corresponding to the event which
                will be set as the preferred event for this superevent. This
                event must be in the same ``category`` as the superevent and
                not already attached to a superevent.
            category (:obj:`str`, optional): superevent category. Allowed
                choices are: 'production', 'test', 'mdc'.
            events (:obj:`str` or :obj:`list[str]`, optional): graceid or list
                of graceids corresponding to events which should be included in
                this superevent. Events must be in the same ``category`` as the
                superevent and not already attached to a superevent.
            labels (:obj:`str` or :obj:`list[str]`, optional): label or list of
                labels which should be attached to this superevent at creation.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:

            >>> g = GraceDb()
            >>> r = g.createSuperevent(1, 2, 3, 'G123456',
            ... category='production', events=['G100', 'G101'],
            ... labels=['EM_READY', 'DQV'])
            >>> r.status_code
            201
        """  # noqa: W605
        # Process label args - convert non-empty strings to list
        # to ensure consistent processing
        if labels:
            if isinstance(labels, six.string_types):
                labels = [labels]
            elif isinstance(labels, (list, tuple)):
                # Validate each entry
                if any([not isinstance(l, six.string_types) for l in labels]):
                    err_msg = "One of the provided labels is not a string"
                    raise TypeError(err_msg)
            else:
                raise TypeError("labels arg is {0}, should be str or list"
                                .format(type(labels)))
            # Check labels against those in database
            for l in labels:
                if l not in self.allowed_labels:
                    raise ValueError(("Label '{0}' does not exist in the "
                                     "database").format(l))
        if events:
            if isinstance(events, six.string_types):
                events = [events]
            elif isinstance(events, (list, tuple)):
                if any([not isinstance(e, six.string_types) for e in events]):
                    err_msg = \
                        "One of the provided event graceids is not a string"
                    raise TypeError(err_msg)
            else:
                raise TypeError("events arg is {0}, should be str or list"
                                .format(type(events)))

        # Validate category, convert to short form if necessary
        if not isinstance(category, six.string_types):
            err_msg = "category arg is {0}, should be a string".format(
                type(category))
            raise TypeError(err_msg)
        category = self._getCode(category, self.superevent_categories)
        # Note: category can be None here as a result of processing
        # by _getCode, which indicates failure to match the provided
        # superevent category
        if not category:
            raise ValueError("category must be one of: {0}".format(
                list(six.itervalues(self.superevent_categories))))

        # Set up request body for POST
        request_body = {
            't_start': t_start,
            't_0': t_0,
            't_end': t_end,
            'preferred_event': preferred_event,
            'category': category,
        }
        if events:
            request_body['events'] = events
        if labels:
            request_body['labels'] = labels

        # Python httplib bug?  unicode link
        uri = self.links['superevents']
        return self.post(uri, data=request_body)

    def updateSuperevent(self, superevent_id, t_start=None, t_0=None,
                         t_end=None, preferred_event=None, em_type=None,
                         time_coinc_far=None, space_coinc_far=None):
        r"""Update a superevent's parameters.

        The same permission restrictions apply as for
        :py:meth:`ligo.gracedb.rest.GraceDb.createSuperevent`. You must provide
        at least one parameter value to update, otherwise the request will be
        rejected for not providing any new information.

        Args:
            superevent_id (str): GraceDB ID of superevent to update
            t_start (:obj:`float`, optional): new t\ :sub:`start` value for
                superevent
            t_0 (:obj:`float`, optional): new t\ :sub:`0` value for superevent
            t_end (:obj:`float`, optional): new t\ :sub:`end` value for
                superevent
            preferred_event (:obj:`str`, optional): graceid corresponding to
                an event which will be set as the new preferred event for this
                superevent. This event must be in the same ``category`` as the
                superevent (``'production'``, ``'test'``, ``'mdc'``) and must
                either already be a part of this superevent, or not be in a
                superevent at all.
            em_type (:obj:`str`, optional): name of coincident EM search
                for "preferred" EM coincident event. Defined by analyst or
                pipeline (e.g., RAVEN). Default is null for new superevents.
            time_coinc_far (:obj:`float`, optional): new value for temporal
                coincident FAR with preferred EM coincident event. Defined by
                analyst or pipeline  (e.g., RAVEN). Default is null for new
                superevents.
            space_coinc_far (:obj:`float`, optional): new value for temporal
                coincident FAR with preferred EM coincident event. Defined by
                analyst or pipeline  (e.g., RAVEN). Default is null for new
                superevents.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = g.updateSuperevent('S181224a', t_start=12, preferred_event=
            ... 'G654321')
            >>> r.status_code
            200
        """  # noqa: W605
        # Make sure that at least one parameter is provided
        if not (t_start or t_0 or t_end or preferred_event
                or em_type or time_coinc_far or space_coinc_far):
            raise ValueError('Provide at least one of t_start, t_0, t_end, '
                             'preferred_event, em_type, time_coinc_far, '
                             'or space_coinc_far')

        request_body = {}
        if t_start is not None:
            request_body['t_start'] = t_start
        if t_0 is not None:
            request_body['t_0'] = t_0
        if t_end is not None:
            request_body['t_end'] = t_end
        if preferred_event is not None:
            request_body['preferred_event'] = preferred_event
        if em_type is not None:
            request_body['em_type'] = em_type
        if time_coinc_far is not None:
            request_body['time_coinc_far'] = time_coinc_far
        if space_coinc_far is not None:
            request_body['space_coinc_far'] = space_coinc_far
        template = self.templates['superevent-detail-template']
        uri = template.format(superevent_id=superevent_id)
        return self.patch(uri, data=request_body)

    def addEventToSuperevent(self, superevent_id, graceid):
        """Add an event to a superevent.

        The event must be in the same category as the superevent and must not
        already be part of a superevent.

        Args:
            superevent_id (str): GraceDB ID of the superevent to which the
                event will be added.
            graceid (str): graceid of the event to add to this superevent.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = addEventToSuperevent('S181224a', 'G123456')
            >>> r.status_code
            201
        """
        request_body = {'event': graceid}
        template = self.templates['superevent-event-list-template']
        uri = template.format(superevent_id=superevent_id)
        return self.post(uri, data=request_body)

    def removeEventFromSuperevent(self, superevent_id, graceid):
        """Remove an event from a superevent.

        The event must already be a part of the superevent.

        Args:
            superevent_id (str): GraceDB ID of the superevent from which the
                event will be removed.
            graceid (str): graceid of the event to remove from this superevent.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = removeEventFromSuperevent('S181224a', 'G123456')
            >>> r.status_code
            204
        """
        template = self.templates['superevent-event-detail-template']
        uri = template.format(superevent_id=superevent_id, graceid=graceid)
        return self.delete(uri)

    def superevent(self, superevent_id):
        """Get information about an individual superevent.

        Args:
            superevent_id (str): GraceDB ID of the superevent.

        Returns:
            :class:`httplib.HTTPResponse`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> superevent = g.superevent('S181224a').json()
        """
        return self.get(self.templates['superevent-detail-template'].format(
            superevent_id=superevent_id))

    def superevents(self, query='', orderby=None, count=None, columns=None,
                    max_results=None):
        """Search for superevents which match a query.

        Information on forming queries is available here:
        https://gracedb.ligo.org/documentation/queries.html

        Args:
            query (:obj:`str`, optional): query string.
            orderby (:obj:`str`, optional): field to order the results by.
            count (:obj:`int`, optional): each generator iteration will yield
                this many objects (default determined by the server).
            columns (:obj:`list[str]`, optional): list of attributes to return
                for each superevent (default: all).
            max_results (:obj:`int`, optional): maximum number of results to
                return (default: all).

        Returns:
            :obj:`Iterator[dict]`

            An iterator which yields individual superevent dictionaries.

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> for s in g.superevents(query='is_gw: True', orderby=
            ... ['-preferred_event'], columns=['superevent_id', 'events']):
            ...     print(s['superevent_id'])
        """
        # If columns is a comma-separated string, split it to a list
        if isinstance(columns, six.string_types):
            columns = columns.split(',')

        # If orderby is a list (should be), convert it to a comma-separated
        # string (that's what the server expects)
        if isinstance(orderby, (list, tuple)):
            orderby = ",".join(orderby)

        # Get URI
        uri = self.links['superevents']

        # Compile URL parameters
        qdict = {}
        if query:
            qdict['query'] = query
        if count is not None:
            qdict['count'] = count
        if orderby:
            qdict['sort'] = orderby
        if qdict:
            uri += "?" + urlencode(qdict)

        # Get superevent information and construct a generator
        n = 0
        while uri:
            response = self.get(uri).json()
            superevents = response.get('superevents', [])
            uri = response.get('links', {}).get('next')

            for superevent in superevents:
                n += 1
                if (max_results is not None and n > max_results):
                    break
                # If columns are specified, only return specific values
                if columns:
                    yield {k: superevent[k] for k in columns}
                else:
                    yield superevent

    def confirm_superevent_as_gw(self, superevent_id):
        """Upgrade a superevent's state to 'confirmed GW'.

        All LIGO-Virgo users can perform this action on test superevents,
        but special permissions are required for production and MDC
        superevents. This action cannot be undone!

        Args:
            superevent_id (str): GraceDB ID of the superevent to confirm as
                as GW.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = confirm_superevent_as_gw('S181224a')
            >>> r.status_code
            204
        """
        template = self.templates['superevent-confirm-as-gw-template']
        uri = template.format(superevent_id=superevent_id)
        return self.post(uri)

    @event_or_superevent
    def files(self, object_id, filename="", *args, **kwargs):
        """Get a list of files or download a file associated with an event or
        superevent.

        If ``filename`` is not provided, get a list of available files
        associated with the event or superevent. If ``filename`` *is*
        provided, download the contents of that file.

        Args:
            object_id (str): event graceid or superevent ID.
            filename (:obj:`str`, optional): name of file to download.

        Returns:
            :class:`requests.models.Response`

            When ``filename`` is not specified, ``response.json()`` contains a
            dict with file basename keys and full file URL values.
            When ``filename`` is specified, use ``response.read()`` to get the
            contents of the file.

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Examples:
            Get a list of files:

            >>> g = GraceDb()
            >>> event_files = g.files('T101383').json()
            >>> for filename in list(event_files):
            ...     # do something
            ...     pass

            Get a file's content:

            >>> outfile = open('my_skymap.png', 'w')
            >>> r = g.files('T101383', 'skymap.png')
            >>> outfile.write(r.content)
            >>> outfile.close()
        """
        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            uri_kwargs = {'superevent_id': object_id}
            if filename:
                # Get specific file
                uri_kwargs['file_name'] = filename
                template = self.templates['superevent-file-detail-template']
            else:
                # Get list of files
                template = self.templates['superevent-file-list-template']
        else:
            template = self.templates['files-template']
            if not filename:
                filename = ""
            uri_kwargs = {'graceid': object_id, 'filename': filename}
        uri = template.format(**uri_kwargs)

        # Make request:
        if filename:
            return self.get_file(uri)
        else:
            return self.get(uri)

    @event_or_superevent
    def logs(self, object_id, log_number=None, *args, **kwargs):
        """Get log entries associated with an event or superevent.

        If ``log_number`` is specified, only a single log message is returned.
        Otherwise, all log messages attached to the event or superevent in
        question are returned.

        Args:
            object_id (str): event graceid or superevent ID.
            log_number (:obj:`int`, optional): ID number (N) of the log
                entry to retrieve.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Examples:

            Get all log messages:

            >>> g = GraceDb()
            >>> response_dict = g.logs('T101383').json()
            >>> print "Num logs = %d" % response_dict['numRows']
            >>> log_list = response_dict['log']
            >>> for log in log_list:
            ...     print log['comment']

        Get a single log message:

            >>> g = GraceDb()
            >>> log_info = g.logs('T101383', 10).json()
        """
        if log_number is not None and not isinstance(log_number, int):
            raise TypeError('log_number should be an int')

        # Set up template and object id
        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            uri_kwargs = {'superevent_id': object_id}
            if log_number:
                template = self.templates['superevent-log-detail-template']
            else:
                template = self.templates['superevent-log-list-template']
        else:
            uri_kwargs = {'graceid': object_id}
            if log_number:
                template = self.templates['event-log-detail-template']
            else:
                template = self.templates['event-log-template']

        if log_number is not None:
            uri_kwargs['N'] = log_number

        uri = template.format(**uri_kwargs)
        return self.get(uri)

    @event_or_superevent
    def writeLog(self, object_id, message, filename=None, filecontents=None,
                 tag_name=[], displayName=[], *args, **kwargs):
        """Create a new log entry associated with an event or superevent.

        Args:
            object_id (str): event graceid or superevent ID.
            message (str): comment to post.
            filename (:obj:`str`, optional): path to file to be uploaded.
                Use ``'-'`` to read from stdin.
            filecontents (:obj:`file`, optional): handler pointing to a file to
                be read and uploaded. If this argument is specified, the
                contents will be saved as ``filename`` on the server.
            tag_name (:obj:`str` or :obj:`list[str]`, optional): tag name or
                list of tag names to be applied to the log message.
            displayName (:obj:`str` or :obj:`list[str]`, optional): tag display
                string or list of display strings for the tag(s) in
                ``tag_name``. If provided, there should be one for each tag.
                Not applicable for tags which already exist on the server.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = g.writeLog('T101383', 'Good stuff.', '/path/to/plot.png',
            ... tag_name='analyst_comments')
            >>> print r.status_code
            201
        """

        # Handle old usage of 'tagname' instead of 'tag_name'
        tagname = kwargs.pop('tagname', None)
        if tagname is not None and not tag_name:
            tag_name = tagname

        # Handle cases where tag_name is a string
        if isinstance(tag_name, str):
            tag_name = [tag_name]
        elif isinstance(tag_name, (tuple, set)):
            tag_name = list(tag_name)
        elif tag_name is None:
            tag_name = []

        # Handle cases where displayName is a string
        if isinstance(displayName, str):
            displayName = [displayName]
        elif isinstance(displayName, (tuple, set)):
            displayName = list(displayName)
        elif displayName is None:
            displayName = []

        # Check displayName length - should be 0 or same as tag_name
        if (displayName and isinstance(tag_name, list)
            and len(displayName) != len(tag_name)):
            raise ValueError("For a list of tags, either provide no display "
                             "names or a display name for each tag")

        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            template = self.templates['superevent-log-list-template']
            uri_kwargs = {'superevent_id': object_id}
        else:
            template = self.templates['event-log-template']
            uri_kwargs = {'graceid': object_id}
        uri = template.format(**uri_kwargs)
        files = None
        if filename:
            if filecontents is None:
                if filename == '-':
                    filename = 'stdin'
                    filecontents = sys.stdin.read()
                else:
                    filecontents = open(filename, 'rb')
            files = {'upload': (os.path.basename(filename),
                                filecontents,
                                get_mimetype(filename))}

        # Set up body of request
        body = {
            'comment': message,
            'tagname': tag_name,
            'displayName': displayName,
        }

        # If files are attached, we have to encode the request body
        # differently, so we convert from a dict to a list of tuples.
        if files:
            body = dict_to_form_encoded(body)

        return self.post(uri, data=body, files=files)

    @event_or_superevent
    def emobservations(self, object_id, emobservation_num=None, *args,
                       **kwargs):
        """Get EM observation data associated with an event or superevent.

        If ``emobservation_num`` is provided, get a single EM observation data
        entry. Otherwise, retrieve all EM observation data associated with the
        event or superevent in question.

        Args:
            object_id (str): event graceid or superevent ID.
            emobservation_num (:obj:`int`, optional): ID number (N) of the EM
                observation to retrieve.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Examples:

            Get a list of EM observations:

            >>> g = GraceDb()
            >>> r = g.emobservations('T101383')
            >>> full_dictionary = r.json()
            >>> emo_list = full_dictionary['observations']

            Get a single EM observation:

            >>> g = GraceDb()
            >>> r = g.emobservations('T101383', 2)
            >>> observation_dict = r.json()
        """
        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            uri_kwargs = {'superevent_id': object_id}
            if emobservation_num:
                template = \
                    self.templates['superevent-emobservation-detail-template']
            else:
                template = \
                    self.templates['superevent-emobservation-list-template']
        else:
            uri_kwargs = {'graceid': object_id}
            if emobservation_num:
                template = self.templates['emobservation-detail-template']
            else:
                template = self.templates['emobservation-list-template']

        if emobservation_num is not None:
            uri_kwargs['N'] = emobservation_num

        uri = template.format(**uri_kwargs)
        return self.get(uri)

    @event_or_superevent
    def writeEMObservation(self, object_id, group, raList, raWidthList,
                           decList, decWidthList, startTimeList, durationList,
                           comment="", *args, **kwargs):
        """Create an EM observation data entry for an event or superevent.

        Args:
            object_id (str): event graceid or superevent ID.
            group (str): name of EM MOU group making the observation.
            raList (:obj:`list[float]`): list of right ascension coordinates
                (degrees).
            raWidthList (:obj:`list[float]` or :obj:`float`): list of right
                ascension measurement widths OR a single number if all
                measurements have the same width (degrees).
            decList (:obj:`list[float]`): list of declination coordinates
                (degrees).
            decWidthList (:obj:`list[float]` or :obj:`float`): list of
                declination measurement widths OR a single number if all
                measurements have the same width (degrees).
            startTimeList (:obj:`list[str]`): list of measurement start times
                in ISO 8601 format and UTC time zone.
            durationList (:obj:`list[float]` or :obj:`float`): list of exposure
                times OR a single number if all measurements have the same
                exposure (seconds).
            comment (:obj:`str`, optional): comment on observation.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = g.writeEMObservation('S190131g', 'ZTF', [1, 2, 3],
            ... [0.1, 0.1, 0.1], [4, 5, 6], 0.2, [],
            ... [10, 9, 11], comment="data uploaded")
            >>> r.status
            201
        """
        # Validate EM group
        if group not in self.em_groups:
            err_msg = "group must be one of {groups}".format(
                groups=", ".join(self.em_groups))
            raise ValueError(err_msg)

        # Argument checking
        num_measurements = len(raList)
        # convert any single number widths or durations into lists
        raWidthList, decWidthList, durationList = \
            [[l] * num_measurements if not isinstance(l, (list, tuple)) else l
                for l in [raWidthList, decWidthList, durationList]]

        # Compare all list lengths
        all_lists = [raList, decList, startTimeList, raWidthList,
                     decWidthList, durationList]
        if not all(map(lambda l: len(l) == num_measurements, all_lists)):
            raise ValueError('raList, decList, startTimeList, raWidthList, '
                             'decWidthList, and durationList should be the '
                             'same length')

        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            template = self.templates['superevent-emobservation-list-template']
            uri_kwargs = {'superevent_id': object_id}
        else:
            template = self.templates['emobservation-list-template']
            uri_kwargs = {'graceid': object_id}
        uri = template.format(**uri_kwargs)

        body = {
            'group': group,
            'ra_list': raList,
            'ra_width_list': raWidthList,
            'dec_list': decList,
            'dec_width_list': decWidthList,
            'start_time_list': startTimeList,
            'duration_list': durationList,
            'comment': comment,
        }
        return self.post(uri, json=body)

    @event_or_superevent
    def labels(self, object_id, label="", *args, **kwargs):
        """Get a label or labels attached to an event or superevent.

        Args:
            object_id (str): event graceid or superevent ID.
            label (:obj:`str`, optional): name of label.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Examples:

            Get a list of labels:

            >>> g = GraceDb()
            >>> label_list = g.labels('T101383').json()['labels']
            >>> for label in label_list:
            ...     print label['name']

            Get a single label:

            >>> g = GraceDb()
            >>> dqv_label = g.labels('T101383', 'DQV').json()
        """
        # Check label name
        if label and label not in self.allowed_labels:
            raise NameError(("Label '{0}' does not exist in the "
                            "database").format(label))

        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            uri_kwargs = {'superevent_id': object_id}
            if label:
                template = self.templates['superevent-label-detail-template']
                uri_kwargs['label_name'] = label
            else:
                template = self.templates['superevent-label-list-template']
        else:
            template = self.templates['event-label-template']
            uri_kwargs = {'graceid': object_id}
            uri_kwargs['label'] = label

        uri = template.format(**uri_kwargs)
        return self.get(uri)

    @event_or_superevent
    def writeLabel(self, object_id, label, *args, **kwargs):
        """Add a label to an event or superevent.

        Args:
            object_id (str): event graceid or superevent ID.
            label (str): label name. Use
                :py:attr:`ligo.gracedb.rest.GraceDb.allowed_labels` to get
                a list of available labels on the server.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = g.writeLabel('T101383', 'DQV')
            >>> r.status_code
            201
        """
        # Check label name
        if label not in self.allowed_labels:
            raise NameError(("Label '{0}' does not exist in the "
                            "database").format(label))

        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            template = self.templates['superevent-label-list-template']
            uri_kwargs = {'superevent_id': object_id}
            request_body = {'name': label}
            uri = template.format(**uri_kwargs)
            return self.post(uri, data=request_body)
        else:
            template = self.templates['event-label-template']
            uri_kwargs = {'graceid': object_id}
            uri_kwargs['label'] = label
            request_body = {}
            uri = template.format(**uri_kwargs)
            return self.put(uri, data=request_body)

    @event_or_superevent
    def removeLabel(self, object_id, label, *args, **kwargs):
        """Remove a label from an event or superevent.

        Args:
            object_id (str): event graceid or superevent ID.
            label (str): label name. The label must be presently applied
                to the event or superevent.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = g.removeLabel('T101383', 'DQV')
            >>> r.status_code
            204
        """
        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            template = self.templates['superevent-label-detail-template']
            uri_kwargs = {'superevent_id': object_id}
            uri_kwargs['label_name'] = label
        else:
            template = self.templates['event-label-template']
            uri_kwargs = {'graceid': object_id}
            uri_kwargs['label'] = label
        uri = template.format(**uri_kwargs)
        return self.delete(uri)

    @event_or_superevent
    def tags(self, object_id, N, *args, **kwargs):
        """Get tags attached to an event or superevent log entry.

        Args:
            object_id (str): event graceid or superevent ID.
            N (int): ID number (N) of the log entry for which to get tags.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> tag_list = g.tags('T101383', 56).json()['tags']
            >>> print "Number of tags for message 56: %d" % len(tag_list)
        """
        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            uri_kwargs = {'superevent_id': object_id}
            template = self.templates['superevent-log-tag-list-template']
        else:
            uri_kwargs = {'graceid': object_id}
            template = self.templates['taglist-template']
        uri_kwargs['N'] = N
        uri = template.format(**uri_kwargs)
        return self.get(uri)

    @event_or_superevent
    def addTag(self, object_id, N, tag_name, displayName=None, *args,
               **kwargs):
        """Add a tag to an event or superevent log entry.

        Args:
            object_id (str): event graceid or superevent ID.
            N (int): ID number of log entry.
            tag_name (str): name of tag to add; tags which don't already exist
                on the server will be created.
            displayName (:obj:`str`, optional): tag display name; only used
                to create new tags which don't already exist.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = g.createTag('T101383', 56, 'sky_loc')
            >>> r.status
            201
        """
        is_superevent = kwargs.pop('is_superevent', False)
        request_body = {}

        # Add displayName to requestBody, if applicable.
        if displayName is not None:
            request_body['displayName'] = displayName

        if is_superevent:
            template = self.templates['superevent-log-tag-list-template']
            uri_kwargs = {'superevent_id': object_id}
            uri_kwargs['N'] = N
            request_body['name'] = tag_name
            uri = template.format(**uri_kwargs)
            return self.post(uri, data=request_body)
        else:
            template = self.templates['tag-template']
            uri_kwargs = {'graceid': object_id, 'tag_name': tag_name}
            uri_kwargs['N'] = N
            uri = template.format(**uri_kwargs)
            return self.put(uri, data=request_body)

    @event_or_superevent
    def removeTag(self, object_id, N, tag_name, *args, **kwargs):
        """Remove a tag from an event or superevent log entry.

        Args:
            object_id (str): event graceid or superevent ID.
            N (int): ID number of log entry.
            tag_name (str): name of tag to add; tags which don't already exist
                on the server will be created.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = g.deleteTag('T101383', 56, 'sky_loc')
            >>> r.status
            200
        """
        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            template = self.templates['superevent-log-tag-detail-template']
            uri_kwargs = {'superevent_id': object_id}
        else:
            template = self.templates['tag-template']
            uri_kwargs = {'graceid': object_id}
        uri_kwargs['N'] = N
        uri_kwargs['tag_name'] = tag_name
        uri = template.format(**uri_kwargs)
        return self.delete(uri)

    def ping(self):
        """Ping the server.

        Returns:
            :class:`requests.models.Response`

            ``response.json()`` contains the information from the API root.

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.
        """
        return self.get(self.links['self'])

    @event_or_superevent
    def voevents(self, object_id, voevent_num=None, *args, **kwargs):
        """Get a VOEvent or list of VOEvents for an event or superevent.

        To get the XML file associated with a VOEvent, use the ``filename``
        in the response JSON with :py:meth:`ligo.gracedb.rest.GraceDb.files`.

        Args:
            object_id (str): event graceid or superevent ID.
            voevent_num (:obj:`int`, optional): ID number (N) of the VOEvent
                to retrieve.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Examples:
            Get a list of VOEvents:

            >>> g = GraceDb()
            >>> r = g.voevents('T101383')
            >>> voevent_list = r.json()['voevents']

            Get a single VOEvent:

            >>> g = GraceDb()
            >>> r = g.voevents('T101383', 2)
            >>> voevent = r.json()
        """
        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            uri_kwargs = {'superevent_id': object_id}
            if voevent_num:
                template = self.templates['superevent-voevent-detail-template']
            else:
                template = self.templates['superevent-voevent-list-template']
        else:
            uri_kwargs = {'graceid': object_id}
            if voevent_num:
                template = self.templates['voevent-detail-template']
            else:
                template = self.templates['voevent-list-template']

        if voevent_num is not None:
            uri_kwargs['N'] = voevent_num

        uri = template.format(**uri_kwargs)
        return self.get(uri)

    @event_or_superevent
    def createVOEvent(self, object_id, voevent_type, skymap_type=None,
                      skymap_filename=None, combined_skymap_filename=None,
                      internal=True, open_alert=False, raven_coinc=False,
                      hardware_inj=False, CoincComment=False, ProbHasNS=None,
                      ProbHasRemnant=None, BNS=None, NSBH=None, BBH=None,
                      Terrestrial=None, MassGap=None, *args, **kwargs):
        r"""Create a new VOEvent.

        Args:
            object_id (str): event graceid or superevent ID.
            voevent_type (str): VOEvent type (choose from ``'preliminary'``,
                ``'initial'``, ``'update'``, and ``'retraction'``).
            skymap_type (:obj:`str`, optional): skymap type. Required for
                VOEvents which include a skymap.
            skymap_filename (:obj:`str`, optional): name of skymap file on
                the GraceDB server (required for 'initial' and 'update' alerts,
                optional for 'preliminary' alerts).
            combined_skymap_file (:obj:`str`, optional): name of combined
                skymap file, if present
            internal (:obj:`bool`, optional): whether the VOEvent should be
                distributed to LIGO-Virgo members only.
            hardware_inj (:obj:`bool`, optional): whether the candidate is a
                hardware injection.
            open_alert (:obj:`bool`, optional): whether the candidate is an
                open alert or not.
            raven_coinc (:obj:`bool`, optional): is VOEvent result of
                coincidence from RAVEN pipeline. Tells GraceDB to look for
                `coinc_far` and `em_type` data models.
            CoincComment (:obj:`bool`, optional): whether the candidate has a
                possible counterpart GRB.
            ProbHasNS (:obj:`float`, optional): probability that at least one
                object in the binary is less than 3 M\ :sub:`sun` (CBC events
                only).
            ProbHasRemnant (:obj:`float`): probability that there is matter in
                the surroundings of the central object (CBC events only).
            BNS (:obj:`float`, optional): probability that the source is a
                binary neutron star merger (CBC events only).
            NSBH (:obj:`float`, optional): probability that the source is a
                neutron star-black hole merger (CBC events only).
            BBH (:obj:`float`, optional): probability that the source is a
                binary black hole merger (CBC events only).
            Terrestrial (:obj:`float`, optional): probability that the source
                is terrestrial (i.e., a background noise fluctuation or a
                glitch) (CBC events only).
            MassGap (:obj:`float`, optional): probability that at least one
                object in the binary has a mass between 3 and 5 M\ :sub:`sun`
                (CBC events only).

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = g.createVOEvent('T101383', 'initial', skymap_type='custom',
            ... skymap_filename='skymap.fits.gz', internal=True, ProbHasNS=0.7,
            ... MassGap=0.4, Terrestrial=0.05)
            >>> r.status
            201
        """  # noqa: W605
        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            template = self.templates['superevent-voevent-list-template']
            uri_kwargs = {'superevent_id': object_id}
        else:
            template = self.templates['voevent-list-template']
            uri_kwargs = {'graceid': object_id}
        uri = template.format(**uri_kwargs)

        # validate voevent_type, convert to short form if necessary
        voevent_type = self._getCode(voevent_type.lower(), self.voevent_types)
        if not voevent_type:
            raise ValueError("voevent_type must be one of: {0}".format(
                ", ".join(list(six.itervalues(self.voevent_types)))))

        # Require skymaps for 'update' and 'initial'
        if voevent_type == 'IN':
            if not skymap_filename:
                raise ValueError("Skymap file is required for 'initial' "
                                 "VOEvents")

        # Construct request body
        body = {
            'voevent_type': voevent_type,
            'internal': internal,
            'open_alert': open_alert,
            'hardware_inj': hardware_inj,
            'CoincComment': CoincComment,
            'raven_coinc': raven_coinc,
        }
        # Add optional args
        if skymap_type is not None:
            body['skymap_type'] = skymap_type
        if skymap_filename is not None:
            body['skymap_filename'] = skymap_filename
        if combined_skymap_filename is not None:
            body['combined_skymap_filename'] = combined_skymap_filename
        if ProbHasNS is not None:
            body['ProbHasNS'] = ProbHasNS
        if ProbHasRemnant is not None:
            body['ProbHasRemnant'] = ProbHasRemnant
        if BNS is not None:
            body['BNS'] = BNS
        if NSBH is not None:
            body['NSBH'] = NSBH
        if BBH is not None:
            body['BBH'] = BBH
        if Terrestrial is not None:
            body['Terrestrial'] = Terrestrial
        if MassGap is not None:
            body['MassGap'] = MassGap

        return self.post(uri, data=body)

    @event_or_superevent
    def permissions(self, object_id, *args, **kwargs):
        """Get a list of permissions for a superevent.

        Only LIGO-Virgo members are allowed to access this information.
        This is not currently implemented for events.

        Args:
            object_id (str): superevent ID.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> g.permissions('S190304bc').json()['permissions']
            [
                {'group': 'group1', 'permission': 'view_superevent'},
                {'group': 'group1', 'permission': 'annotate_superevent'},
                {'group': 'group2', 'permission': 'view_superevent'},
            ]
        """
        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            template = self.templates[
                'superevent-permission-list-template']
            uri_kwargs = {'superevent_id': object_id}
        else:
            raise NotImplementedError('Not implemented for events')

        uri = template.format(**uri_kwargs)
        return self.get(uri)

    @event_or_superevent
    def modify_permissions(self, object_id, action, *args, **kwargs):
        """Expose or hide a superevent to/from the public.

        This action requires special server-side permissions. It is not
        yet implemented for events.

        Args:
            object_id (str): superevent ID.
            action (str): ``'expose'`` or ``'hide'``.

        Returns:
            :class:`httplib.HTTPResponse`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:

            Expose a superevent:

            >>> g = GraceDb()
            >>> g.modify_permissions('S190304bc', 'expose').json()
            [
                {'group': 'group1', 'permission': 'view_superevent'},
                {'group': 'group1', 'permission': 'annotate_superevent'},
                {'group': 'group2', 'permission': 'view_superevent'},
            ]

            Hide a superevent:

            >>> g.modify_permissions('S190304bc', 'hide').json()
            []
        """

        if (action not in ['expose', 'hide']):
            raise ValueError('action should be \'expose\' or \'hide\'')

        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            template = self.templates[
                'superevent-permission-modify-template']
            uri_kwargs = {'superevent_id': object_id}
        else:
            raise NotImplementedError('Not implemented for events')

        uri = template.format(**uri_kwargs)

        body = {'action': action}
        return self.post(uri, data=body)

    def _signoff_helper(self, object_id, action, template, uri_kwargs,
                        signoff_type=None, instrument=None, status=None,
                        comment=None):
        # NOTE: uri_kwargs should already have the graceid or superevent_id
        # in it

        # Validate args
        if signoff_type:
            signoff_type = self._getCode(signoff_type, self.signoff_types)
            if not signoff_type:
                raise ValueError("signoff_type must be one of: {0}".format(
                    ", ".join(self.signoff_types)))
        if instrument:
            instrument = self._getCode(instrument, self.instruments)
            if not instrument:
                raise ValueError("instrument must be one of: {0}".format(
                    ", ".join(self.instruments)))
        if status:
            status = self._getCode(status, self.signoff_statuses)
            if not status:
                raise ValueError("status must be one of: {0}".format(
                    ", ".join(self.signoff_statuses)))
        if signoff_type == 'OP' and not instrument:
            raise ValueError('Operator signoffs require an instrument')

        # Get HTTP method and args
        body = {}
        if (action == 'create'):
            http_method = "POST"
            body['signoff_type'] = signoff_type
            body['instrument'] = instrument
            body['comment'] = comment
            body['status'] = status
        elif (action == 'update'):
            http_method = "PATCH"
            uri_kwargs['typeinst'] = signoff_type + instrument
            if comment is not None:
                body['comment'] = comment
            if status is not None:
                body['status'] = status
        elif (action == 'get'):
            http_method = "GET"
            if signoff_type is not None:
                uri_kwargs['typeinst'] = signoff_type + instrument
        elif (action == 'delete'):
            http_method = "DELETE"
            uri_kwargs['typeinst'] = signoff_type + instrument
        else:
            raise ValueError("action should be 'create', 'update', "
                             "'get', or 'delete'")
        uri = template.format(**uri_kwargs)

        # Get http method
        method = getattr(self, http_method.lower())
        if body:
            response = method(uri, data=body)
        else:
            response = method(uri)

        return response

    @event_or_superevent
    def signoffs(self, object_id, signoff_type=None, instrument='', *args,
                 **kwargs):
        """Get a signoff or list of signoffs for a superevent.

        This action is not yet implemented for events.

        Args:
            object_id (str): superevent ID.
            signoff_type (str): signoff type. Choices are: ``'OP'`` or
                ``'operator'`` (operator signoff), or ``'ADV'`` or
                ``'advocate'`` (advocate signoff).
            instrument (:obj:`str`, optional): instrument abbreviation
                (``'H1'``, ``'L1'``, etc.) for operator signoffs. Leave blank
                for advocate signoffs.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:

            Get a list of signoffs:

            >>> g = GraceDb()
            >>> signoffs = g.signoffs('S190221z').json()['signoffs']

            Get a single signoff:

            >>> signoff = g.signoffs('S190221z', 'H1').json()
        """
        # Get URI template
        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            if signoff_type is not None:
                template = self.templates['superevent-signoff-detail-template']
            else:
                template = self.templates['superevent-signoff-list-template']
            uri_kwargs = {'superevent_id': object_id}
        else:
            raise NotImplementedError('Not yet implemented for events')

        return self._signoff_helper(
            object_id, 'get', template, uri_kwargs,
            signoff_type=signoff_type, instrument=instrument
        )

    @event_or_superevent
    def create_signoff(self, object_id, signoff_type, status, comment,
                       instrument='', *args, **kwargs):
        """Create a superevent signoff.

        This action requires special server-side permissions. You must be in
        the control room of an IFO to perform operator signoffs or in the
        'em_advocates' group to perform advocate signoffs. This action is not
        yet implemented for events.

        Args:
            object_id (str): superevent ID.
            signoff_type (str): signoff type. Choices are: ``'OP'`` or
                ``'operator'`` (operator signoff), or ``'ADV'`` or
                ``'advocate'`` (advocate signoff).
            status (str): signoff status (``'OK'`` or ``'NO'``).
            comment (str): comment on the signoff.
            instrument (:obj:`str`, optional): instrument abbreviation
                (``'H1'``, ``'L1'``, etc.) for operator signoffs. Leave blank
                for advocate signoffs.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = g.create_signoff('S190102p', 'OP', 'OK',
            ... 'LHO looks good.', instrument='H1')
            >>> r.status_code
            201
        """

        # Get URI template
        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            template = self.templates['superevent-signoff-list-template']
            uri_kwargs = {'superevent_id': object_id}
        else:
            raise NotImplementedError('Not yet implemented for events')

        return self._signoff_helper(
            object_id, 'create', template, uri_kwargs,
            signoff_type=signoff_type, instrument=instrument, status=status,
            comment=comment
        )

    @event_or_superevent
    def update_signoff(self, object_id, signoff_type, status=None,
                       comment=None, instrument='', *args, **kwargs):
        """Update a superevent signoff.

        This action requires the same permissions as
        :py:meth:`ligo.gracedb.rest.GraceDb.create_signoff`.

        Args:
            object_id (str): superevent ID.
            signoff_type (str): signoff type. Choices are: ``'OP'`` or
                ``'operator'`` (operator signoff), or ``'ADV'`` or
                ``'advocate'`` (advocate signoff).
            status (:obj:`str`, optional): signoff status (if changed).
            comment (:obj:`str`, optional): comment on the signoff
                (if changed).
            instrument (:obj:`str`, optional): instrument abbreviation
                (``'H1'``, ``'L1'``, etc.) for operator signoffs. Leave blank
                for advocate signoffs.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = g.update_signoff('S190102p', 'OP', status='NO',
            ... comment='IFO status was actually bad.', instrument='H1')
            >>> r.status_code
            200
        """
        # This will make a PATCH request

        # Either status or comment must be included - otherwise the user
        # is not updating anything
        if not (status or comment):
            raise ValueError("Provide at least one of 'status' or 'comment'")

        # Get URI template
        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            template = self.templates['superevent-signoff-detail-template']
            uri_kwargs = {'superevent_id': object_id}
        else:
            raise NotImplementedError('Not yet implemented for events')

        return self._signoff_helper(
            object_id, 'update', template, uri_kwargs,
            signoff_type=signoff_type, instrument=instrument, status=status,
            comment=comment
        )

    @event_or_superevent
    def delete_signoff(self, object_id, signoff_type, instrument='', *args,
                       **kwargs):
        """Delete a superevent signoff.

        This action requires the same permissions as ``create_signoff()``.

        Args:
            object_id (str): superevent ID.
            signoff_type (str): signoff type. Choices are: ``'OP'`` or
                ``'operator'`` (operator signoff), or ``'ADV'`` or
                ``'advocate'`` (advocate signoff).
            instrument (:obj:`str`, optional): instrument abbreviation
                (``'H1'``, ``'L1'``, etc.) for operator signoffs. Leave blank
                for advocate signoffs.

        Returns:
            :class:`requests.models.Response`

        Raises:
            ligo.gracedb.exceptions.HTTPError: if the response has a status
                code >= 400.

        Example:
            >>> g = GraceDb()
            >>> r = g.delete_signoff('S190102p', 'OP', instrument='H1')
            >>> r.status_code
            204
        """
        # Get URI template
        is_superevent = kwargs.pop('is_superevent', False)
        if is_superevent:
            template = self.templates['superevent-signoff-detail-template']
            uri_kwargs = {'superevent_id': object_id}
        else:
            raise NotImplementedError('Not yet implemented for events')

        return self._signoff_helper(
            object_id, 'delete', template, uri_kwargs,
            signoff_type=signoff_type, instrument=instrument
        )



import pytest

# Apply module-level mark
pytestmark = pytest.mark.integration


@pytest.mark.parametrize("obj_type", ['event', 'superevent'])
def test_labels(client, create_obj, obj_type):
    # Create event or superevent
    obj, obj_id = create_obj(obj_type)

    # Add a label
    label1 = 'DQV'
    response = client.writeLabel(obj_id, label1)
    assert response.status_code == 201
    data = response.json()
    # NOTE: events API returns empty response - should be fixed
    if obj_type == 'superevent':
        assert data['name'] == label1
    elif obj_type == 'event':
        assert data == '{}'

    # Add another label
    label2 = 'INJ'
    response = client.writeLabel(obj_id, label2)
    assert response.status_code == 201
    data = response.json()
    # NOTE: events API returns empty response - should be fixed
    if obj_type == 'superevent':
        assert data['name'] == label2
    elif obj_type == 'event':
        assert data == '{}'

    # Get list of labels
    response = client.labels(obj_id)
    assert response.status_code == 200
    data = response.json()
    obj_labels = [l['name'] for l in data['labels']]
    assert len(obj_labels) == 2
    assert label1 in obj_labels
    assert label2 in obj_labels

    # Pull down an individual label
    response = client.labels(obj_id, label2)
    assert response.status_code == 200
    data = response.json()
    assert data['name'] == label2

    # Remove a label
    response = client.removeLabel(obj_id, label1)
    assert response.status_code == 204

    # Get event and double check
    if obj_type == 'event':
        method = 'event'
    elif obj_type == 'superevent':
        method = 'superevent'
    response = getattr(client, method)(obj_id)
    assert response.status_code == 200
    data = response.json()
    assert len(data['labels']) == 1
    assert label1 not in data['labels']
    assert label2 in data['labels']



# Tests for level 2 subcommand help
#  Ex: 'gracedb add event'
import pytest
try:
    from unittest import mock
except ImportError:
    import mock

from .test_level1_commands import COMMANDS_TEST_DATA

# Apply module-level mark
pytestmark = pytest.mark.cli


###############################################################################
# Test data ###################################################################
###############################################################################
SUBCOMMANDS_TEST_DATA = [(t[0], sc) for t in COMMANDS_TEST_DATA for sc in t[1]]


###############################################################################
# Parametrized tests ##########################################################
###############################################################################
@pytest.mark.parametrize("cmd,subcommand", SUBCOMMANDS_TEST_DATA)
def test_subcommand_help(CLI, cmd, subcommand):
    """Test level 2 subcommand help functionality"""
    command_line_cmd = '{cmd} {sc} --help'.format(cmd=cmd, sc=subcommand)

    print_help_func_str = \
        'ligo.gracedb.cli.parsers.argparse.ArgumentParser.print_help'
    with mock.patch(print_help_func_str) as mock_print_help:
        with pytest.raises(SystemExit) as excinfo:
            CLI(command_line_cmd.split(' '))

    # Make sure print_help is called once
    assert mock_print_help.call_count == 1

    # Check exit code
    assert excinfo.value.code == 0


@pytest.mark.parametrize("cmd,subcommand", SUBCOMMANDS_TEST_DATA)
def test_subcommand_help_content(CLI, cmd, subcommand):
    """Check level 2 subcommand help content"""

    # Create an instance of the subcommand
    sc_instance = CLI.subcommand_dict[cmd](parent_prog=CLI.parser.prog)

    # Get help dialogue
    help_text = sc_instance.parser.format_help()

    # Check usage
    assert help_text.startswith('usage: {prog} {sc}'.format(
        prog=CLI.parser.prog, sc=cmd))

    # Check description - have to replace newlines with spaces since the
    # help text is specially formatted
    desc = getattr(sc_instance, 'long_description', sc_instance.description)
    assert desc.replace('\n', ' ') in help_text.replace('\n', ' ')



import pytest

from ligo.gracedb.utils import handle_str_or_list_arg


def test_str_or_list_arg_processor_str():
    data = 'test_str'
    out = handle_str_or_list_arg(data, 'arg_name')
    assert out == [data]


@pytest.mark.parametrize("data", [None, {}, (), [], False])
def test_str_or_list_arg_processor_falsey(data):
    out = handle_str_or_list_arg(data, 'arg_name')
    assert out == data


def test_str_or_list_arg_processor_list():
    data = [1, 2, 3]
    out = handle_str_or_list_arg(data, 'arg_name')
    assert out == data


@pytest.mark.parametrize("data", [{'a': 1}, (1, 2,), 1, 3.4, True])
def test_str_or_list_arg_processor_bad_data(data):
    arg_name = 'arg_name'
    err_msg = "{an} arg is {at}, should be str or list".format(
        an=arg_name, at=type(data))
    with pytest.raises(TypeError, match=err_msg):
        handle_str_or_list_arg(data, arg_name)



import itertools
import pytest
try:
    from unittest import mock
except ImportError:  # py < 3
    import mock

from ligo.gracedb.rest import GraceDb
from ligo.gracedb.cert import check_certificate_expiration


def test_x509_cert_load():
    """Test loading of X.509 certificate during client instantiation"""
    # Set up cert and key files
    cert_file = '/tmp/cert_file'
    key_file = '/tmp/key_file'

    load_cert_func = 'ligo.gracedb.cert.load_certificate'
    with mock.patch(load_cert_func) as mock_load_cert:
        # Initialize client
        g = GraceDb(cred=(cert_file, key_file))

    # Check credentials
    assert len(g.cert) == 2
    assert g.auth_type == 'x509'
    assert g.cert[0] == cert_file
    assert g.cert[1] == key_file
    assert mock_load_cert.called_once()


@pytest.mark.parametrize("reload_buffer", [86400, 10])
def test_x509_cert_expiration(reload_buffer, x509_cert):
    """Test X.509 certificate expiration check"""

    # Check if certificate is expired (should have 3600 second lifetime)
    # compared to reload_buffer
    expired = check_certificate_expiration(x509_cert,
                                           reload_buffer=reload_buffer)
    if reload_buffer > 3600:
        assert expired is True
    else:
        assert expired is False

# AEP: test depreciated.
# def test_x509_cert_autoload_in_expiration_check():

# AEP: test depreciated. 'load_certificate" decoupled from client class
# and does not depend on auth_type.
# def test_load_certificate_with_auth_type_not_x509():

# AEP: test depeciated. 'check_certificate_expiration' decoupled from
# client class and does not check auth_type.
# def test_check_certificate_with_auth_type_not_x509():


# All possible combinations of True/False for the three variables
RELOAD_TEST_DATA = list(itertools.product((True, False), repeat=3))
@pytest.mark.skip(reason="tested online, figure out mocking in pytest'")
@pytest.mark.parametrize("force_noauth,reload_cert,cert_expired",  # noqa: E302
                         RELOAD_TEST_DATA)
def test_reloading_feature(force_noauth, reload_cert, cert_expired):
    # Set up cert and key files
    cert_file = '/tmp/cert_file'
    key_file = '/tmp/key_file'

    set_up_conn_func = \
        'ligo.gracedb.adapter.GraceDbCertHTTPSConnection.connect'
    get_conn_func = \
        'ligo.gracedb.adapter.GraceDbCertHTTPSConnectionPool._get_conn'
    load_cert_func = 'ligo.gracedb.cert.load_certificate'
    cert_expire_func = \
        'ligo.gracedb.cert.check_certificate_expiration'
    with mock.patch(get_conn_func), \
         mock.patch(set_up_conn_func) as mock_set_up_conn, \
         mock.patch(load_cert_func) as mock_load_cert, \
         mock.patch(cert_expire_func) as mock_cert_expire:  # noqa: E127

        # Set return value for mock_cert_expire
        mock_cert_expire.return_value = cert_expired

        # Initialize client
        g = GraceDb(
            cred=(cert_file, key_file),
            reload_certificate=reload_cert,
            force_noauth=force_noauth
        )

        # Try to make a request
        g.get("https://fakeurl.com")

        # Compile number of times which we expect certain functions to be
        # called
        check_cert_call_count = 0
        set_up_conn_call_count = 0
        load_cert_call_count = int(not force_noauth)

        if g.auth_type == 'x509' and reload_cert:
            check_cert_call_count += 1
            if cert_expired:
                load_cert_call_count += 1
                set_up_conn_call_count += 1

        # Compare to actual results
        assert mock_load_cert.call_count == load_cert_call_count
        assert mock_cert_expire.call_count == check_cert_call_count
        assert mock_set_up_conn.call_count == set_up_conn_call_count



import pytest

# Set module-level marks
pytestmark = pytest.mark.integration


def test_ping(client):
    response = client.ping()
    assert response.status_code == 200


@pytest.mark.parametrize(
    "resource",
    ['api_versions', 'server_version', 'links', 'templates', 'groups',
     'pipelines', 'searches', 'allowed_labels', 'superevent_categories',
     'em_groups', 'voevent_types', 'signoff_types', 'signoff_statuses',
     'instruments']
)
def test_api_root_content(client, resource):
    assert getattr(client, resource) is not None



import os

import pytest

# Apply module-level marks
pytestmark = pytest.mark.integration


@pytest.mark.parametrize("obj_type", ['event', 'superevent'])
def test_file_upload_and_retrieval(client, test_data_dir, create_obj,
                                   obj_type):
    # Create event or superevent
    obj, obj_id = create_obj(obj_type)

    # Create a log with a file upload
    comment = 'test file upload'
    test_file = os.path.join(test_data_dir, 'test_file.txt')
    response = client.writeLog(obj_id, comment, filename=test_file)
    assert response.status_code == 201
    data = response.json()
    assert data['comment'] == comment
    assert data['filename'] == os.path.basename(test_file)

    # Check file list for obj
    response = client.files(obj_id)
    assert response.status_code == 200
    data = response.json()
    assert len(data) >= 1

    # Download the file and compare it
    response = client.files(obj_id, os.path.basename(test_file))
    assert response.status_code == 200
    file_from_server = response.read()
    with open(test_file, 'rb') as f:
        file_from_disk = f.read()
    assert file_from_server == file_from_disk


@pytest.mark.parametrize("obj_type", ['event', 'superevent'])
def test_upload_file_multiple_versions(client, test_data_dir, create_obj,
                                       obj_type):
    # Create event or superevent
    obj, obj_id = create_obj(obj_type)

    # Create a log with a file upload
    file_name = 'new_file_test.txt'
    comment = 'test file upload'
    fc1 = 'file 1 data'
    response = client.writeLog(obj_id, comment, filename=file_name,
                               filecontents=fc1)
    assert response.status_code == 201
    data = response.json()
    assert data['comment'] == comment
    assert data['filename'] == file_name
    assert data['file_version'] == 0

    # Create another log using the same file name
    fc2 = 'file 2 data'
    response = client.writeLog(obj_id, comment, filename=file_name,
                               filecontents=fc2)
    assert response.status_code == 201
    data = response.json()
    assert data['comment'] == comment
    assert data['filename'] == file_name
    assert data['file_version'] == 1

    # Check the attached files
    response = client.files(obj_id)
    assert response.status_code == 200
    data = response.json()
    # There should be at least the two files we uploaded and the symlink to
    # the latest one.
    assert len(data) >= 3
    for version in ['', ',0', ',1']:
        assert (file_name + version) in data

    # Download all files and compare the contents
    file1_data = client.files(obj_id, file_name + ',0').read()
    file2_data = client.files(obj_id, file_name + ',1').read()
    file_data = client.files(obj_id, file_name).read()
    assert file1_data.decode() == fc1
    assert file2_data.decode() == fc2
    assert file2_data == file_data
    assert file1_data != file2_data


@pytest.mark.parametrize("obj_type", ['event', 'superevent'])
def test_upload_large_file(client, test_data_dir, create_obj, obj_type):
    # Create event or superevent
    obj, obj_id = create_obj(obj_type)

    # Upload a large file (2 MB)
    filename = os.path.join(test_data_dir, 'big.data')
    response = client.writeLog(obj_id, "Large file upload test",
                               filename=filename)
    assert response.status_code == 201
    data = response.json()
    assert data['filename'] == os.path.basename(filename)

    # Download file and compare to file on disk
    # Download the file and compare it
    response = client.files(obj_id, os.path.basename(filename))
    assert response.status_code == 200
    file_from_server = response.read()
    with open(filename, 'rb') as f:
        file_from_disk = f.read()
    assert file_from_server == file_from_disk


@pytest.mark.parametrize("obj_type", ['event', 'superevent'])
def test_upload_binary_file(client, test_data_dir, create_obj, obj_type):
    # Create event or superevent
    obj, obj_id = create_obj(obj_type)

    # Upload a binary file
    filename = os.path.join(test_data_dir, 'upload.data.gz')
    response = client.writeLog(obj_id, "Binary file upload test",
                               filename=filename)
    assert response.status_code == 201
    data = response.json()
    assert data['filename'] == os.path.basename(filename)



# Tests for subcommands below 'create'
#  Ex: 'gracedb create emobservation'
import pytest
import shlex
try:
    from unittest import mock
except ImportError:
    import mock

# Apply module-level mark
pytestmark = pytest.mark.cli


###############################################################################
# Tests of individual subcommands #############################################
###############################################################################
CMD_DATA = ['events', 'superevents']
@pytest.mark.parametrize("command", CMD_DATA)  # noqa: E302
def test_search_subcommand(CLI, command):
    """Test search events and superevents subcommands"""
    query = 'fake query'

    # Compile command
    cols = ['col1', 'col2.test']
    max_results = 4
    cmd = 'search {cmd} "{query}" --columns={cols} --max-results={N}' \
        .format(cmd=command, query=query, cols=','.join(cols), N=max_results)

    # Mock return value
    rv = [
        {'col1': 11, 'col2': {'test': 21}},
        {'col1': 12, 'col2': {'test': 22}},
        {'col1': 13, 'col2': {'test': 23}},
        {'col1': 14, 'col2': {'test': 24}},
    ]

    func = 'ligo.gracedb.rest.GraceDb.{cmd}'.format(cmd=command)
    with mock.patch(func) as mock_cli_func:
        mock_cli_func.return_value = iter(rv)
        response = CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (query,)
    assert cli_kwargs == {'max_results': max_results}

    # Check response
    lines = response.split('\n')
    assert lines[0] == '#{headers}'.format(headers="\t".join(cols))
    for i, line in enumerate(lines[1:]):
        assert line == '\t'.join(
            [str(rv[i]['col1']), str(rv[i]['col2']['test'])])
